# -*- coding: utf-8 -*-
"""
Created on Tue Jan 30 12:35:43 2024

@author: Andreas Gebauer
"""

import time
import shutil
import os
import numpy as np
import scipy.optimize
import scipy.interpolate
import AG_pulses_m8_website as pulses
import scipy.signal
import pandas as pd

#%%

def read_Teraflash_1file(file):
    f = open(file)
    lines = f.readlines()
    f.close()
    
    t = []
    S = []
    ref = []
    for line in lines:
        if line.startswith('Time_abs/ps,') or line.startswith('#') or line.startswith('loop') or line.startswith('sample') or line.startswith('angle') or line.startswith('average'):
            continue
        s = line.split()
        t.append(float(s[0][:-1]))
        S.append(float(s[1][:-1]))
        ref.append(float(s[2]))
    return np.array(t), np.array(S), np.array(ref)




def read_Teraflash_2files(ref_file, sam_file):
    f = open(ref_file)
    lines = f.readlines()
    f.close()
    t_ref = []
    Et_ref = []
    for line in lines:
        if line.startswith(('%', 'Time_abs/ps,', '#', 'loop', 'sample', 'angle', 'average')):
            continue
        s = line.strip().split(',')
        try:
            t_ref.append(float(s[0].replace(',', '').replace(' ', '').strip()))  # Fix trailing commas
            Et_ref.append(float(s[1].replace(',', '').replace(' ', '').strip()))
        except ValueError:
            print(f"Skipping line in {ref_file} due to format error: {line.strip()}")

    t_ref = np.array(t_ref)
    Et_ref = np.array(Et_ref)

    f = open(sam_file)
    lines = f.readlines()
    f.close()
    t_sam = []
    Et_sam = []
    for line in lines:
        if line.startswith(('%', 'Time_abs/ps,', '#', 'loop', 'sample', 'angle', 'average')):
            continue
        s = line.strip().split(',')
        try:
            t_sam.append(float(s[0].replace(',', '').replace(' ', '').strip()))  # Fix trailing commas
            Et_sam.append(float(s[1].replace(',', '').replace(' ', '').strip()))
        except ValueError:
            print(f"Skipping line in {sam_file} due to format error: {line.strip()}")

    t_sam = np.array(t_sam)
    Et_sam = np.array(Et_sam)

    if abs(min(t_sam) - min(t_ref)) > 1e-14 or abs(max(t_sam) - max(t_ref)) > 1e-14 or abs(len(t_sam) - len(t_ref)) > 1e-14:
        print(f'read_Teraflash_2files: reference ({ref_file}) and sample ({sam_file}) have different delay axes. t_sam is returned.')

    return t_sam, Et_sam, Et_ref






#%%
"""
Generate time domain data for demonstration
"""

def transmission(n, k, d, f, N_echo):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    T = 4*N / (N + 1)**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * (N - 1) * d))
    etalon = 1
    for m in range(1,N_echo+1):
        etalon += (((N - 1)/(N + 1))**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * N * 2 * d)))**m
    return T*etalon

def lorentzian_oscillator_model(f, f0, tau, fp):
    epsilon = 1 + (2*np.pi*fp)**2 / (((f0*2*np.pi)**2 - (f*2*np.pi)**2) - 1j * 2*np.pi*f/tau)
    n = np.sqrt(1/2 * np.sqrt(epsilon.real**2 + epsilon.imag**2) + epsilon.real/2)
    k = np.sqrt(1/2 * np.sqrt(epsilon.real**2 + epsilon.imag**2) - epsilon.real/2)
    return n, k

def generate_timedomain_data_lorentzian_oscillator(input_pulse, Lorentzian_oscillator):
    # inpult_pulse is dictionary that contains:
    #   t_min: float, minimum time delay [ps]
    #   t_max: float, maximum time delay [ps]
    #   dt: float, time delay step [ps]
    #   f0: float, input pulse central frequency [THz]
    #   T: float, input pulse pulse duration [ps]
    #   chirp: float, input pulse chirp [ps^2]
    #   CEP: float, input pulse carrier envelope phase [rad]
    # Lorentzian_oscillator is a dictionary that contains:
    #   f_res: float, resonance frequency [THz]
    #   T_res: float, resonance lifetime [ps]
    #   fp: float, plasma frequency
    #   d: float, thickness of sample [m]
    #   N_echo: int, number of echo pulses to include in transmission function
    # Output:  
    #   t: time delay axis [ps]
    #   Et_in: reference pulse 
    #   Et_out: sample pulse
    #   f: w/2pi, frequency axis [THz]
    #   n: real part of refractive index
    #   k: imaginary part of refractive index
    t = np.linspace(input_pulse['t_min'], input_pulse['t_max'], int((input_pulse['t_max'] - input_pulse['t_min']) / input_pulse['dt']) + 1)
    Et_in = pulses.gaussian_chirped_pulse(t = t, omega = 2*np.pi * input_pulse['f0'], T = input_pulse['T'], chirp = input_pulse['chirp'], CEP = input_pulse['CEP'], E0 = 1) # input pulse in time-domain
    w, Ew_in = pulses.t2w_transform(t = t, Et = Et_in) # w is angular frequency (omega) from FFT, Ew_in = input pulse in frequency domain
    n, k = lorentzian_oscillator_model(f = w / (2*np.pi), f0 = Lorentzian_oscillator['f_res'], tau = Lorentzian_oscillator['T_res'], fp = Lorentzian_oscillator['fp'])
    trans = np.nan_to_num(transmission(n = n, k = k, d = Lorentzian_oscillator['d'], f = w / (2*np.pi) * 1e12, N_echo = int(Lorentzian_oscillator['N_echo'])))
    t, Et_out = pulses.w2t_transform(w = w, Ew = trans * Ew_in)
    return t, Et_in.real, Et_out.real, w/(2*np.pi), n, k
    
#%%
"""
model = 'single_layer'

exact optimization method for single layer sample, e.g. substrates:

          (air / vacuum)|   layer  | (air / vacuum) 
 incident light         |<----d--->|      transmitted light
----------------------->|    n,k   |-------------------------------->
"""

def transmission_0echo(n, k, d, f):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    T = 4*N / (N + 1)**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * (N - 1) * d))
    return T

def transmission_1echo(n, k, d, f):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    T = 4*N * (N - 1)**2 / (N + 1)**4 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * (3 * N - 1) * d))
    return T

def thickness_from_nk_no_echo(f, n, k, T0, d_estimate):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    d = d_estimate + c0 / (2 * np.pi * f * (n - 1)) * (np.angle(4 * N / (N + 1)**2) - np.angle(T0 * np.exp(1j * 2*np.pi*f / c0 * (n - 1) * d_estimate)))
    return d

def root_T0abs_n_single_layer(n, k, f, T0, d0):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    trans0_abs_log = -2*np.pi*f / c0 * k * thickness_from_nk_no_echo(f = f, n = n, k = k, T0 = T0, d_estimate = d0) + np.nan_to_num(np.log(np.nan_to_num(abs(4 * N / (N + 1)**2))))
    return trans0_abs_log - np.log(abs(T0))

def root_T1abs_n_single_layer(n, k, f, T0, T1, d0):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    trans1_abs_log = -3 * 2*np.pi*f / c0 * k * thickness_from_nk_no_echo(f = f, n = n, k = k, T0 = T0, d_estimate = d0) + np.nan_to_num(np.log(np.nan_to_num(abs(4 * N * (N - 1)**2 / (N + 1)**4))))
    return trans1_abs_log - np.log(abs(T1))

def root_k_single_layer(klog, f, T0, T1, d0, nmin, nmax):
    n_T0abs = scipy.optimize.brentq(root_T0abs_n_single_layer, nmin, nmax, args = (np.exp(klog), f, T0, d0))
    n_T1abs = scipy.optimize.brentq(root_T1abs_n_single_layer, nmin, nmax, args = (np.exp(klog), f, T0, T1, d0))
    return n_T0abs- n_T1abs

def mini_k_single_layer(klog, f, T0, T1, d0, nmin, nmax):
    n_T0abs = scipy.optimize.brentq(root_T0abs_n_single_layer, nmin, nmax, args = (np.exp(klog), f, T0, d0))
    n_T1abs = scipy.optimize.brentq(root_T1abs_n_single_layer, nmin, nmax, args = (np.exp(klog), f, T0, T1, d0))
    return np.log(abs(n_T0abs- n_T1abs))

def root_T0abs_k_single_layer(klog, n, f, T0, d0):
    c0 = 299792458
    k = np.exp(klog)
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    trans0_abs_log = -2*np.pi*f / c0 * k * thickness_from_nk_no_echo(f = f, n = n, k = k, T0 = T0, d_estimate = d0) + np.nan_to_num(np.log(np.nan_to_num(abs(4 * N / (N + 1)**2))))
    return trans0_abs_log - np.log(abs(T0))

def root_T1abs_k_single_layer(klog, n, f, T0, T1, d0):
    c0 = 299792458
    k = np.exp(klog)
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    trans1_abs_log = -3 * 2*np.pi*f / c0 * k * thickness_from_nk_no_echo(f = f, n = n, k = k, T0 = T0, d_estimate = d0) + np.nan_to_num(np.log(np.nan_to_num(abs(4 * N * (N - 1)**2 / (N + 1)**4))))
    return trans1_abs_log - np.log(abs(T1))

def root_n_single_layer(n, f, T0, T1, d0, klogmin, klogmax):
    k_T0abs = scipy.optimize.brentq(root_T0abs_k_single_layer, klogmin, klogmax, args = (n, f, T0, d0))
    k_T1abs = scipy.optimize.brentq(root_T1abs_k_single_layer, klogmin, klogmax, args = (n, f, T0, T1, d0))
    return k_T0abs- k_T1abs

def mini_n_single_layer(n, f, T0, T1, d0, klogmin, klogmax):
    k_T0abs = scipy.optimize.brentq(root_T0abs_k_single_layer, klogmin, klogmax, args = (n, f, T0, d0))
    k_T1abs = scipy.optimize.brentq(root_T1abs_k_single_layer, klogmin, klogmax, args = (n, f, T0, T1, d0))
    return np.log(abs(k_T0abs- k_T1abs))


def root_T0abs_x1_single_layer_fixed_thickness(x1, x2, f, T0, d):
    # x1 and x2 are auxilary coordinates that rotate the contours of the root by 60° to avoid vertical contours
    n = (x1 + np.sqrt(3) * x2)/2
    k = (-np.sqrt(3) * x1 + x2)/2
    c0 = 299792458
    logabs0 = np.log(4 * np.sqrt(n**2 + k**2) / ((n + 1)**2 + k**2)) - 2*np.pi*f / c0 * k * d
    return logabs0 - np.log(abs(T0))

def root_T0phase_x1_single_layer_fixed_thickness(x1, x2, f, T0, d):
    # x1 and x2 are auxilary coordinates that rotate the contours of the root by 60° to avoid vertical contours
    n = (x1 + np.sqrt(3) * x2)/2
    k = (-np.sqrt(3) * x1 + x2)/2
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    phase0 = np.angle(4 * N / (N + 1)**2) - (np.mod((2*np.pi*f / c0 * (n - 1) * d), 2*np.pi)) + 2*np.pi
    return np.mod(np.angle(T0), 2*np.pi) - phase0
    
def root_x2_single_layer_fixed_thickness(x2, f, T0, d, x1min, x1max):
    x1_T0abs = scipy.optimize.brentq(root_T0abs_x1_single_layer_fixed_thickness, x1min, x1max, args = (x2, f, T0, d))
    x1_T0phase = scipy.optimize.brentq(root_T0phase_x1_single_layer_fixed_thickness, x1min, x1max, args = (x2, f, T0, d))
    return x1_T0abs - x1_T0phase

def mini_x2_single_layer_fixed_thickness(x2, f, T0, d, x1min, x1max):
    x1_T0abs = scipy.optimize.brentq(root_T0abs_x1_single_layer_fixed_thickness, x1min, x1max, args = (x2, f, T0, d))
    x1_T0phase = scipy.optimize.brentq(root_T0phase_x1_single_layer_fixed_thickness, x1min, x1max, args = (x2, f, T0, d))
    return np.log(abs(x1_T0abs - x1_T0phase))

#%%
"""
model = 'single_layer_doublepass'

exact optimization method for double-pass sample, e.g. cryostat windows:

          (air / vacuum)|   layer  | air/vac        |   layer  | (air vacuum) 
 incident light         |<----d--->|<---arbitrary-->|<----d--->|      transmitted light
----------------------->|    n,k   |                |    n,k   |-------------------------------->
"""

def transmission_0echo_doublepass(n, k, d, f):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    T = (4*N / (N + 1)**2)**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * (N - 1) * d * 2))
    return T

def transmission_1echo_doublepass(n, k, d, f):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    T = 2 * (4*N / (N + 1)**2)**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * (N - 1) * d * 2)) * (((N - 1)/(N + 1))**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * N * 2 * d)))
    return T

def thickness_from_nk_no_echo_doublepass(f, n, k, T0, d_estimate):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    d = d_estimate + c0 / (4 * np.pi * f * (n - 1)) * (np.angle((4 * N / (N + 1)**2)**2) - np.angle(T0 * np.exp(1j * 2*np.pi*f / c0 * (n - 1) * 2 * d_estimate)))
    return d

def root_T0abs_n_single_layer_doublepass(n, k, f, T0, d0):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    trans0_abs_log = np.nan_to_num(np.log(np.nan_to_num(abs(4 * N / (N + 1)**2)**2))) - 2*np.pi*f / c0 * k * 2 * d0 - k / (n - 1) * (np.nan_to_num(np.angle((4 * N / (N + 1)**2)**2)) - np.nan_to_num(np.angle(T0 * np.exp(1j * 2*np.pi*f / c0 * (n - 1) * 2 * d0))))
    return trans0_abs_log - np.log(abs(T0))

def root_T1abs_n_single_layer_doublepass(n, k, f, T0, T1, d0):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    trans1_abs_log = np.nan_to_num(np.log(np.nan_to_num(2 * abs(4 * N / (N + 1)**2)**2) * abs((N - 1)/(N + 1))**2)) - 2*np.pi*f / c0 * k * 4 * d0 - 2* k / (n - 1) * (np.nan_to_num(np.angle((4 * N / (N + 1)**2)**2)) - np.nan_to_num(np.angle(T0 * np.exp(1j * 2*np.pi*f / c0 * (n - 1) * 2 * d0))))
    return trans1_abs_log - np.log(abs(T1))

def root_k_single_layer_doublepass(klog, f, T0, T1, d0, nmin, nmax):
    n_T0abs = scipy.optimize.brentq(root_T0abs_n_single_layer_doublepass, nmin, nmax, args = (np.exp(klog), f, T0, d0))
    n_T1abs = scipy.optimize.brentq(root_T1abs_n_single_layer_doublepass, nmin, nmax, args = (np.exp(klog), f, T0, T1, d0))
    return n_T0abs- n_T1abs

def mini_k_single_layer_doublepass(klog, f, T0, T1, d0, nmin, nmax):
    n_T0abs = scipy.optimize.brentq(root_T0abs_n_single_layer_doublepass, nmin, nmax, args = (np.exp(klog), f, T0, d0))
    n_T1abs = scipy.optimize.brentq(root_T1abs_n_single_layer_doublepass, nmin, nmax, args = (np.exp(klog), f, T0, T1, d0))
    return np.log(abs(n_T0abs- n_T1abs))

def root_T0abs_x1_single_layer_doublepass_fixed_thickness(x1, x2, f, T0, d):
    #n = (x2 + x1)/np.sqrt(2)
    #k = (x2 - x1)/np.sqrt(2) 
    n = (x1 + np.sqrt(3) * x2)/2
    k = (-np.sqrt(3) * x1 + x2)/2
    c0 = 299792458
    logabs0 = np.log(16 * (n**2 + k**2) / ((n + 1)**2 + k**2)**2) - 2*np.pi*f / c0 * k * d * 2
    return logabs0 - np.log(abs(T0))

def root_T0phase_x1_single_layer_doublepass_fixed_thickness(x1, x2, f, T0, d):
    #n = (x2 + x1)/np.sqrt(2)
    #k = (x2 - x1)/np.sqrt(2) 
    n = (x1 + np.sqrt(3) * x2)/2
    k = (-np.sqrt(3) * x1 + x2)/2
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    phase0 = np.angle((4 * N / (N + 1)**2)**2) - (np.mod((2*np.pi*f / c0 * (n - 1) * 2 * d), 2*np.pi)) + 2*np.pi
    return np.mod(np.angle(T0), 2*np.pi) - phase0
    
def root_x2_single_layer_doublepass_fixed_thickness(x2, f, T0, d, x1min, x1max):
    x1_T0abs = scipy.optimize.brentq(root_T0abs_x1_single_layer_doublepass_fixed_thickness, x1min, x1max, args = (x2, f, T0, d))
    x1_T0phase = scipy.optimize.brentq(root_T0phase_x1_single_layer_doublepass_fixed_thickness, x1min, x1max, args = (x2, f, T0, d))
    return x1_T0abs - x1_T0phase

def mini_x2_single_layer_doublepass_fixed_thickness(x2, f, T0, d, x1min, x1max):
    x1_T0abs = scipy.optimize.brentq(root_T0abs_x1_single_layer_doublepass_fixed_thickness, x1min, x1max, args = (x2, f, T0, d))
    x1_T0phase = scipy.optimize.brentq(root_T0phase_x1_single_layer_doublepass_fixed_thickness, x1min, x1max, args = (x2, f, T0, d))
    return np.log(abs(x1_T0abs - x1_T0phase))


#%%
"""
model = 'single_layer_cryostat_echo'
exact optimization method for single layer sample in cryostat, where cryostat window echoes overlap temporally with sample echo:

          (air / vacuum)|   window | air/vac   |  sample  | air/vac   |   window | (air vacuum) 
 incident light         |<---dw--->|<---arb.-->|<---ds--->|<---arb.-->|<---dw--->|      transmitted light
----------------------->|  nw, kw  |           |   n, k   |           |  nw, kw  |---------------------->
"""

def transmission_1echo_single_layer_cryostat_echo(n, k, d, f, nw, kw, dw):
    c0 = 299792458
    N = n - 1j*k * np.where(np.sign(f) == 0, 1, np.sign(f))
    Nw = nw - 1j*kw * np.where(np.sign(f) == 0, 1, np.sign(f))
    #T = 4*N / (N + 1)**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * (N - 1) * d)) * (((N - 1)/(N + 1))**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * N * 2 * d))) * (1 + 1/2 * ((N - 1) * (Nw + 1) / ((N + 1) * (Nw - 1)))**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * 2 * (N * d - Nw * dw))))
    T = 4 * N / (N + 1)**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * (N - 1) * d)) * (((N - 1) / (N + 1))**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * N * 2 * d)) + 2 * ((Nw - 1) / (Nw + 1))**2 * np.nan_to_num(np.exp(-2j * np.pi * f / c0 * Nw * 2 * dw)))
    return T

def root_T1abs_n_single_layer_cryostat_echo(n, k, f, T0, T1, d0, nw, kw, dw):
    d = thickness_from_nk_no_echo(f = f, n = n, k = k, T0 = T0, d_estimate = d0)
    trans1_abs_log = np.nan_to_num(np.log(np.nan_to_num(abs(transmission_1echo_single_layer_cryostat_echo(n = n, k = k, d = d, f = f, nw = nw, kw = kw, dw = dw)))))
    return trans1_abs_log - np.log(abs(T1))

def root_k_single_layer_cryostat_echo(klog, f, T0, T1, d0, nw, kw, dw, nmin, nmax):
    n_T0abs = scipy.optimize.brentq(root_T0abs_n_single_layer, nmin, nmax, args = (np.exp(klog), f, T0, d0))
    n_T1abs = scipy.optimize.brentq(root_T1abs_n_single_layer_cryostat_echo, nmin, nmax, args = (np.exp(klog), f, T0, T1, d0, nw, kw, dw))
    return n_T0abs- n_T1abs

def mini_k_single_layer_cryostat_echo(klog, f, T0, T1, d0, nw, kw, dw, nmin, nmax):
    n_T0abs = scipy.optimize.brentq(root_T0abs_n_single_layer, nmin, nmax, args = (np.exp(klog), f, T0, d0))
    n_T1abs = scipy.optimize.brentq(root_T1abs_n_single_layer_cryostat_echo, nmin, nmax, args = (np.exp(klog), f, T0, T1, d0, nw, kw, dw))
    return np.log(abs(n_T0abs- n_T1abs))

#%%
"""
exact optimization method
"""

import matplotlib.pyplot as plt
def get_edges_for_pcolormesh(x_axis):
    dx = np.diff(x_axis)
    X = x_axis[1:] - dx/2
    X = np.insert(X, 0, x_axis[0] - dx[0]/2)
    X = np.append(X, x_axis[-1] + dx[-1]/2)
    return X

def determine_optical_constants_in_interval(f, T0, T1, d_estimate, n_lim, klog_lim, model, model_parameters):
    # determine n, k and sample thickness for various models by numerically solving the exact transmission equations
    # f (float): Frequency in THz
    # T0 (complex): Transmission of main pulse
    # T1 (complex): Transmission of 1st echo pulse
    # d_estimate (float): Estimated thickness of sample in m (needs to be estimated within one wavelength/n)
    # n_lim (list of 2 floats): Root finding limits for real part of refractive index
    # klog_lim (list of 2 floats): Root finding limits for the logarithm of the imaginary part of refractive index
    # model (string): Model name
    # model_parameters (dictionary): Model parameters, elements depend on model
    # returns: 
    #   n (float): real part of refractive index 
    #   k (float): imaginary part of refractive index
    #   d (float): thickness of sample in m
    #   optimization (string): returns "bisection" if bisection method was successful, "minimum" if contours do not cross in the investigated interval or "no_solution" if no appropriate interval was found at all.
    f = f * 1e12
    if model == 'single_layer':
        # check if there is a sign change between nmin and nmax
        cond_nmin = root_n_single_layer(n = min(n_lim), f = f, T0 = T0, T1 = T1, d0 = d_estimate, klogmin = min(klog_lim), klogmax = max(klog_lim))
        cond_nmax = root_n_single_layer(n = max(n_lim), f = f, T0 = T0, T1 = T1, d0 = d_estimate, klogmin = min(klog_lim), klogmax = max(klog_lim))
        if cond_nmin * cond_nmax <= 0:
            # sign change in n-interval: Apply bisection method (more precise)
            n = scipy.optimize.brentq(root_n_single_layer, min(n_lim), max(n_lim), args = (f, T0, T1, d_estimate, min(klog_lim), max(klog_lim)))
            optimization = 'bisection'
        else:
            # no sign change in k-interval: Apply minimization method (less precise)
            n = scipy.optimize.minimize_scalar(mini_n_single_layer, bounds = (min(n_lim), max(n_lim)), args = (f, T0, T1, d_estimate, min(klog_lim), max(klog_lim)), method = 'Bounded', options = {'xatol': 1e-16}).x
            optimization = 'minimum'
        # determine k for optimal n:
        klog_T0abs = scipy.optimize.brentq(root_T0abs_k_single_layer, min(klog_lim), max(klog_lim), args = (n, f, T0, d_estimate))
        klog_T1abs = scipy.optimize.brentq(root_T1abs_k_single_layer, min(klog_lim), max(klog_lim), args = (n, f, T0, T1, d_estimate))
        k = np.exp((klog_T0abs + klog_T1abs) / 2)
        # determine thickness for optimal n and k:
        d = thickness_from_nk_no_echo(f = f, n = n, k = k, T0 = T0, d_estimate = d_estimate)
    elif model == 'single_layer_doublepass':
        # check if there is a sign change between kmin and kmax
        cond_kmin = root_k_single_layer_doublepass(klog = min(klog_lim), f = f, T0 = T0, T1 = T1, d0 = d_estimate, nmin = min(n_lim), nmax = max(n_lim))
        cond_kmax = root_k_single_layer_doublepass(klog = max(klog_lim), f = f, T0 = T0, T1 = T1, d0 = d_estimate, nmin = min(n_lim), nmax = max(n_lim))
        if cond_kmin * cond_kmax <= 0:
            # sign change in k-interval: Apply bisection method (more precise)
            k = np.exp(scipy.optimize.brentq(root_k_single_layer_doublepass, min(klog_lim), max(klog_lim), args = (f, T0, T1, d_estimate, min(n_lim), max(n_lim))))
            optimization = 'bisection'
        else:
            # no sign change in k-interval: Apply minimization method (less precise)
            k = np.exp(scipy.optimize.minimize_scalar(mini_k_single_layer_doublepass, bounds = (min(klog_lim), max(klog_lim)), args = (f, T0, T1, d_estimate, min(n_lim), max(n_lim)), method = 'Bounded', options = {'xatol': 1e-16}).x)
            optimization = 'minimum'
        # determine n for optimal k:
        n_T0abs = scipy.optimize.brentq(root_T0abs_n_single_layer_doublepass, min(n_lim), max(n_lim), args = (k, f, T0, d_estimate))
        n_T1abs = scipy.optimize.brentq(root_T1abs_n_single_layer_doublepass, min(n_lim), max(n_lim), args = (k, f, T0, T1, d_estimate))
        n = (n_T0abs + n_T1abs) / 2
        # determine thickness for optimal n and k:
        d = thickness_from_nk_no_echo_doublepass(f = f, n = n, k = k, T0 = T0, d_estimate = d_estimate)
    elif model == 'single_layer_cryostat_echo':
        nw_int = scipy.interpolate.interp1d(model_parameters['fw'], model_parameters['nw'], kind = 'linear', bounds_error = False, fill_value = 'extrapolate')
        nw = nw_int(f)
        kw_int = scipy.interpolate.interp1d(model_parameters['fw'], model_parameters['kw'], kind = 'linear', bounds_error = False, fill_value = 'extrapolate')
        kw = kw_int(f)
    
        # check if there is a sign change between kmin and kmax
        cond_kmin = root_k_single_layer_cryostat_echo(klog = min(klog_lim), f = f, T0 = T0, T1 = T1, d0 = d_estimate, nw = nw, kw = kw, dw = model_parameters['dw'], nmin = min(n_lim), nmax = max(n_lim))
        cond_kmax = root_k_single_layer_cryostat_echo(klog = max(klog_lim), f = f, T0 = T0, T1 = T1, d0 = d_estimate, nw = nw, kw = kw, dw = model_parameters['dw'], nmin = min(n_lim), nmax = max(n_lim))
        if cond_kmin * cond_kmax <= 0:
            # sign change in k-interval: Apply bisection method (more precise)
            k = np.exp(scipy.optimize.brentq(root_k_single_layer_cryostat_echo, min(klog_lim), max(klog_lim), args = (f, T0, T1, d_estimate, nw, kw, model_parameters['dw'], min(n_lim), max(n_lim))))
            optimization = 'bisection'
        else:
            # no sign change in k-interval: Apply minimization method (less precise)
            k = np.exp(scipy.optimize.minimize_scalar(mini_k_single_layer_cryostat_echo, bounds = (min(klog_lim), max(klog_lim)), args = (f, T0, T1, d_estimate, nw, kw, model_parameters['dw'], min(n_lim), max(n_lim)), method = 'Bounded', options = {'xatol': 1e-16}).x)
            optimization = 'minimum'
        # determine n for optimal k:
        n_T0abs = scipy.optimize.brentq(root_T0abs_n_single_layer, min(n_lim), max(n_lim), args = (k, f, T0, d_estimate))
        n_T1abs = scipy.optimize.brentq(root_T1abs_n_single_layer_cryostat_echo, min(n_lim), max(n_lim), args = (k, f, T0, T1, d_estimate, nw, kw, model_parameters['dw']))
        n = (n_T0abs + n_T1abs) / 2
        # determine thickness for optimal n and k:
        d = thickness_from_nk_no_echo(f = f, n = n, k = k, T0 = T0, d_estimate = d_estimate)
    else:
        raise ValueError('determine_optical_constants_in_interval: model %s is not available' % model)
    return n, k, d, optimization


def plot_root_condition_map(n_axis, klog_axis, root0, root1, n0, k0, n1, k1, n0_new, k0_new, n1_new, k1_new, n_intersect, k_intersect, nmin, nmax, klogmin, klogmax, title):
    # plot root maps, contours, intersections and limits for debugging
    """
    plt.clf()
    fig, ax = plt.subplots(1, 2, sharey= True)
    im = ax[0].pcolorfast(get_edges_for_pcolormesh(klog_axis/np.log(10)), get_edges_for_pcolormesh(n_axis), np.transpose(np.sign(root0)), cmap = 'bwr', vmin = -1, vmax = 1)
    ax[0].contour(klog_axis/np.log(10), n_axis, np.transpose(root0), levels = [0], colors = 'k', linestyles = '-')
    ax[0].contour(klog_axis/np.log(10), n_axis, np.transpose(root1), levels = [0], colors = 'k', linestyles = '--')
    if len(k0) != 0 and len(n0) != 0:
        ax[0].plot(k0/np.log(10), n0, 'y.')
        ax[0].plot(k0_new/np.log(10), n0_new, 'y-')
    if len(k1) != 0 and len(n1) != 0:
        ax[0].plot(k1/np.log(10), n1, 'g.')
        ax[0].plot(k1_new/np.log(10), n1_new, 'g-')
    if n_intersect != None and k_intersect != None:
        ax[0].plot(k_intersect/np.log(10), n_intersect, 'kx')
    rect = plt.Rectangle((klogmin/np.log(10), nmin), klogmax/np.log(10) - klogmin/np.log(10), nmax - nmin, facecolor = None, edgecolor = 'k', fill = False, linewidth = 0.5)
    ax[0].add_patch(rect)
    ax[0].set_title(title)
    im = ax[1].pcolorfast(get_edges_for_pcolormesh(klog_axis/np.log(10)), get_edges_for_pcolormesh(n_axis), np.transpose(np.sign(root1)), cmap = 'bwr', vmin = -1, vmax = 1)
    ax[1].contour(klog_axis/np.log(10), n_axis, np.transpose(root0), levels = [0], colors = 'k', linestyles = '-')
    ax[1].contour(klog_axis/np.log(10), n_axis, np.transpose(root1), levels = [0], colors = 'k', linestyles = '--')
    if len(k0) != 0 and len(n0) != 0:
        ax[1].plot(k0/np.log(10), n0, 'y.')
        ax[1].plot(k0_new/np.log(10), n0_new, 'y-')
    if len(k1) != 0 and len(n1) != 0:
        ax[1].plot(k1/np.log(10), n1, 'g.')
        ax[1].plot(k1_new/np.log(10), n1_new, 'g-')
    if n_intersect != None and k_intersect != None:
        ax[1].plot(k_intersect/np.log(10), n_intersect, 'kx')
    rect = plt.Rectangle((klogmin/np.log(10), nmin), klogmax/np.log(10) - klogmin/np.log(10), nmax - nmin, facecolor = None, edgecolor = 'k', fill = False, linewidth = 0.5)
    ax[1].add_patch(rect)  
    ax[0].set_xlim(-8, 0)
    ax[1].set_xlim(-8, 0)
    plt.show()
    """
    plt.clf()
    fig, ax = plt.subplots(1, 2, sharey= True)
    im = ax[0].pcolorfast(get_edges_for_pcolormesh(klog_axis/np.log(10)), get_edges_for_pcolormesh(n_axis), np.transpose(np.sign(root0)), cmap = 'bwr', vmin = -1, vmax = 1)
    if len(k0) != 0 and len(n0) != 0:
        ax[0].plot(k0/np.log(10), n0, 'y.')
        ax[0].plot(k0_new/np.log(10), n0_new, 'y-')
    if len(k1) != 0 and len(n1) != 0:
        ax[0].plot(k1/np.log(10), n1, 'g.')
        ax[0].plot(k1_new/np.log(10), n1_new, 'g-')
    ax[0].contour(klog_axis/np.log(10), n_axis, np.transpose(root0), levels = [0], colors = 'k', linestyles = '-')
    ax[0].contour(klog_axis/np.log(10), n_axis, np.transpose(root1), levels = [0], colors = 'k', linestyles = '--')
    if n_intersect != None and k_intersect != None:
        ax[0].plot(k_intersect/np.log(10), n_intersect, 'kx')
    rect = plt.Rectangle((klogmin/np.log(10), nmin), klogmax/np.log(10) - klogmin/np.log(10), nmax - nmin, facecolor = None, edgecolor = 'k', fill = False, linewidth = 0.5)
    ax[0].add_patch(rect)
    ax[0].set_title(title)
    im = ax[1].pcolorfast(get_edges_for_pcolormesh(klog_axis/np.log(10)), get_edges_for_pcolormesh(n_axis), np.transpose(np.sign(root1)), cmap = 'bwr', vmin = -1, vmax = 1)
    if len(k0) != 0 and len(n0) != 0:
        ax[1].plot(k0/np.log(10), n0, 'y.')
        ax[1].plot(k0_new/np.log(10), n0_new, 'y-')
    if len(k1) != 0 and len(n1) != 0:
        ax[1].plot(k1/np.log(10), n1, 'g.')
        ax[1].plot(k1_new/np.log(10), n1_new, 'g-')
    ax[1].contour(klog_axis/np.log(10), n_axis, np.transpose(root0), levels = [0], colors = 'k', linestyles = '-')
    ax[1].contour(klog_axis/np.log(10), n_axis, np.transpose(root1), levels = [0], colors = 'k', linestyles = '--')
    if n_intersect != None and k_intersect != None:
        ax[1].plot(k_intersect/np.log(10), n_intersect, 'kx')
    rect = plt.Rectangle((klogmin/np.log(10), nmin), klogmax/np.log(10) - klogmin/np.log(10), nmax - nmin, facecolor = None, edgecolor = 'k', fill = False, linewidth = 0.5)
    ax[1].add_patch(rect)  
    #ax[0].set_xlim(klogmin/np.log(10) * 1.1, klogmax/np.log(10) * 0.9)
    #ax[1].set_xlim(klogmin/np.log(10) * 1.1, klogmax/np.log(10) * 0.9)
    ax[0].set_xlim(-2.8, -2.2)
    ax[1].set_xlim(-2.8, -2.2)
    ax[0].set_ylim(1.8, 2.2)
    ax[1].set_ylim(1.8, 2.2)
    plt.show() 
    return

def find_intersect(x1, y1, x2, y2):
    intersect_index = []
    for i in range(len(x2)):
        j = np.argmin(abs(x2[i] - x1))
        if (y1[j-1] - y2[i-1]) * (y1[j] - y2[i]) <= 0 and (i != 0):
            intersect_index.append(j)
    intersect_index = np.unique(np.array(intersect_index), return_index = False)
    intersect_x = x1[intersect_index[0]]
    intersect_y = y1[intersect_index[0]]
    return intersect_x, intersect_y, intersect_index[0]

def find_new_boundaries(f_next, T0_next, T1_next, d_estimate, n_lim, klog_lim, model, model_parameters):
    # Finds neighbouring roots and poles to estimate limits of the parameter space for n and k at the next frequency
    # f_next (float): Frequency in THz (next data point)
    # T0_next (complex): Transmission of main pulse (next data point)
    # T1_next (complex): Transmission of 1st echo pulse (next data point)
    # d_estimate (float): Estimated thickness in m if determine_thickness = True and assumed thickness in m if determine_thickness = False
    # n_prev (float): Real part of refractive index of previous frequency data point
    # n_lim (list of two floats): Estimate limits of real part of refractive index
    # klog_lim (list of two floats): Estimate limits of natural logarithm of imaginary part of refractive index
    # model (string): Model name
    # model_parameters (dictionary): Model parameters, elements depend on model
    # returns:
    # n_lim (list of 2 floats): New root finding limits for real part of refractive index
    # klog_lim (list of 2 floats): New root finding limits for logarithm of imaginary part of refractive index      
    # no_solution (list of booleans): True if it is not possible to find bisection in both roots
    
    """
    new contour-finding has not yet been implemented for models single_layer_doublepass and single_layer_cryostat_echo
    """    
    
    
    f_next = 1e12 * f_next
    # calculate n,k-landscapes for bisection condition
    if model == 'single_layer':
        n_axis = np.linspace(min(n_lim), max(n_lim), 256)
        klog_axis = np.linspace(max([min(klog_lim), np.log(1e-12)]), max(klog_lim), 256)
        NN, KK = np.meshgrid(n_axis, klog_axis)
        root0 = root_T0abs_k_single_layer(klog = KK, n = NN, f = f_next, T0 = T0_next, d0 = d_estimate)
        root1 = root_T1abs_k_single_layer(klog = KK, n = NN, f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate)
    elif model == 'single_layer_doublepass':
        n_axis = np.linspace(min(n_lim), max(n_lim), 256)
        klog_axis = np.linspace(max([min(klog_lim), np.log(1e-12)]), max(klog_lim), 256)
        NN, KK = np.meshgrid(n_axis, klog_axis)
        root0 = root_T0abs_n_single_layer_doublepass(n = NN, k = np.exp(KK), f = f_next, T0 = T0_next, d0 = d_estimate)
        root1 = root_T1abs_n_single_layer_doublepass(n = NN, k = np.exp(KK), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate)
    elif model == 'single_layer_cryostat_echo':
        nw_int = scipy.interpolate.interp1d(model_parameters['fw'], model_parameters['nw'], kind = 'linear', bounds_error = False, fill_value = 'extrapolate')
        nw = nw_int(f_next)
        kw_int = scipy.interpolate.interp1d(model_parameters['fw'], model_parameters['kw'], kind = 'linear', bounds_error = False, fill_value = 'extrapolate')
        kw = kw_int(f_next)
        
        n_axis = np.linspace(min(n_lim), max(n_lim), 256)
        klog_axis = np.linspace(max([min(klog_lim), np.log(1e-12)]), max(klog_lim), 256)
        NN, KK = np.meshgrid(n_axis, klog_axis)
        root0 = root_T0abs_n_single_layer(n = NN, k = np.exp(KK), f = f_next, T0 = T0_next, d0 = d_estimate)
        root1 = root_T1abs_n_single_layer_cryostat_echo(n = NN, k = np.exp(KK), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate, nw = nw, kw = kw, dw = model_parameters['dw'])
    else:
        raise ValueError('find_new_boundaries (cond_n): model %s is not available' % model)
    
    # find zero contour in root0 and root1
    fig_temp, ax_temp = plt.subplots() # create a temporary figure for contour calculation
    contour_T0 = np.transpose(ax_temp.contour(KK, NN, root0, 0).allsegs[1][0]) # calculate contour for T0 equation and extract unique coordinates
    k0 = contour_T0[0]
    n0 = contour_T0[1]
    inds = n0.argsort()
    k0 = k0[inds]
    n0 = n0[inds]
    n0, inds = np.unique(n0, return_index = True)
    k0 = k0[inds]
    contour_T1 = np.transpose(ax_temp.contour(KK, NN, root1, 0).allsegs[1][0]) # calculate contour for T1 equation and extract unique coordinates
    k1 = contour_T1[0]
    n1 = contour_T1[1]
    inds = n1.argsort()
    k1 = k1[inds]
    n1 = n1[inds]
    n1, inds = np.unique(n1, return_index = True)
    k1 = k1[inds]
    ax_temp.cla() # close temporary figures and axes to save RAM space
    plt.close(fig_temp)
    if len(n0) > 1 and len(n1) > 1:
        # zero contours exist in both, root0 and root1.
        # estimate intersection point of contours via interpolation
        n1_new = n1[np.nonzero(np.logical_and(n1 >= min(n_lim), n1 <= max(n_lim)))]
        k1_new = k1[np.nonzero(np.logical_and(n1 >= min(n_lim), n1 <= max(n_lim)))]
        k0_int = scipy.interpolate.interp1d(x = n0, y = k0, kind = 'linear', bounds_error = False, fill_value = 'extrapolate')
        k0_new = k0_int(n1_new)
        n0_new = n1_new
        index_intersect = np.argmin(abs(k0_new - k1_new))
        n_intersect = n1_new[index_intersect]
        k_intersect = k0_int(n_intersect)
        
        nmax_candidates = [n_intersect + 1, max(n0)-1e-2, max(n1)-1e-2]
        nmin_candidates = [n_intersect - 1, min(n_lim), min(n0)+1e-2, min(n1)+1e-2]
        
        dk0dn1 = np.nan_to_num(np.gradient(k0_new) / np.gradient(n1_new))
        peaks = scipy.signal.find_peaks(abs(dk0dn1), prominence = 0.5)
        if len(peaks[0]) > 0:
            for i in range(len(peaks[0])):
                if n1_new[peaks[0][i]] > n_intersect:
                    nmax_candidates.append(n1_new[peaks[0][i]])
                else:
                    nmin_candidates.append(n1_new[peaks[0][i]])
        
        # make sure that root0 and root1 have exactly one sign change between nmin_try and nmax_try at klogmin_try and klogmax_try, respectively
        nmin_try = max(nmin_candidates)
        nmax_try = min(nmax_candidates)
        klogmin_try = min(klog_lim)
        klogmax_try = max(klog_lim)
        control_T0_nmin = 1
        control_T0_nmax = 1
        control_T1_nmin = 1
        control_T1_nmax = 1
        dn = 1e-3
        dk = 1e-2
        while (control_T0_nmin >= 0 or control_T0_nmax >= 0 or control_T1_nmin >= 0 or control_T1_nmax >= 0) and (klogmax_try - klogmin_try <= (max(klog_lim) - min(klog_lim))):
            if model == 'single_layer':
                if nmax_try - nmin_try > 2*dn:
                    nmin_try = nmin_try - dn
                    nmax_try = nmax_try + dn
                else:
                    nmin_try = nmin_try
                    nmax_try = nmax_try
                if klogmax_try - klogmin_try >= 2*dk:    
                    klogmin_try = klogmin_try - dk
                    klogmax_try = klogmax_try + dk
                else:
                    klogmin_try = klogmin_try
                    klogmax_try = klogmax_try
                root_T0_kmin_nmin = root_T0abs_k_single_layer(klog = klogmin_try, n = max([nmin_try, 1.05]), f = f_next, T0 = T0_next, d0 = d_estimate)
                root_T0_kmin_nmax = root_T0abs_k_single_layer(klog = klogmin_try, n = nmax_try, f = f_next, T0 = T0_next, d0 = d_estimate)
                root_T0_kmax_nmin = root_T0abs_k_single_layer(klog = klogmax_try, n = max([nmin_try, 1.05]), f = f_next, T0 = T0_next, d0 = d_estimate) 
                root_T0_kmax_nmax = root_T0abs_k_single_layer(klog = klogmax_try, n = nmax_try, f = f_next, T0 = T0_next, d0 = d_estimate)
                root_T1_kmin_nmin = root_T1abs_k_single_layer(klog = klogmin_try, n = max([nmin_try, 1.05]), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate) 
                root_T1_kmin_nmax = root_T1abs_k_single_layer(klog = klogmin_try, n = nmax_try, f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate)
                root_T1_kmax_nmin = root_T1abs_k_single_layer(klog = klogmax_try, n = max([nmin_try, 1.05]), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate) 
                root_T1_kmax_nmax = root_T1abs_k_single_layer(klog = klogmax_try, n = nmax_try, f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate)
                control_T0_nmin = root_T0_kmin_nmin * root_T0_kmax_nmin
                control_T0_nmax = root_T0_kmin_nmax * root_T0_kmax_nmax
                control_T1_nmin = root_T1_kmin_nmin * root_T1_kmax_nmin
                control_T1_nmax = root_T1_kmin_nmax * root_T1_kmax_nmax
            elif model == 'single_layer_doublepass':
                nmin_try = nmin_try + 0.01
                nmax_try = nmax_try - 0.01
                klogmin_try = klogmin_try + 0.01
                klogmax_try = klogmax_try - 0.01
                control_T0_kmin = root_T0abs_n_single_layer_doublepass(n = max([nmin_try, 1.05]), k = np.exp(klogmin_try), f = f_next, T0 = T0_next, d0 = d_estimate) * root_T0abs_n_single_layer_doublepass(n = nmax_try, k = np.exp(klogmin_try), f = f_next, T0 = T0_next, d0 = d_estimate)
                control_T0_kmax = root_T0abs_n_single_layer_doublepass(n = max([nmin_try, 1.05]), k = np.exp(klogmax_try), f = f_next, T0 = T0_next, d0 = d_estimate) * root_T0abs_n_single_layer_doublepass(n = nmax_try, k = np.exp(klogmax_try), f = f_next, T0 = T0_next, d0 = d_estimate)
                control_T1_kmin = root_T1abs_n_single_layer_doublepass(n = max([nmin_try, 1.05]), k = np.exp(klogmin_try), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate) * root_T1abs_n_single_layer_doublepass(n = nmax_try, k = np.exp(klogmin_try), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate)
                control_T1_kmax = root_T1abs_n_single_layer_doublepass(n = max([nmin_try, 1.05]), k = np.exp(klogmax_try), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate) * root_T1abs_n_single_layer_doublepass(n = nmax_try, k = np.exp(klogmax_try), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate)
            elif model == 'single_layer_cryostat_echo':
                nmin_try = nmin_try + 0.01
                nmax_try = nmax_try - 0.01
                klogmin_try = klogmin_try + 0.01
                klogmax_try = klogmax_try - 0.01
                control_T0_kmin = root_T0abs_n_single_layer(n = max([nmin_try, 1.05]), k = np.exp(klogmin_try), f = f_next, T0 = T0_next, d0 = d_estimate) * root_T0abs_n_single_layer(n = nmax_try, k = np.exp(klogmin_try), f = f_next, T0 = T0_next, d0 = d_estimate)
                control_T0_kmax = root_T0abs_n_single_layer(n = max([nmin_try, 1.05]), k = np.exp(klogmax_try), f = f_next, T0 = T0_next, d0 = d_estimate) * root_T0abs_n_single_layer(n = nmax_try, k = np.exp(klogmax_try), f = f_next, T0 = T0_next, d0 = d_estimate)
                control_T1_kmin = root_T1abs_n_single_layer_cryostat_echo(n = max([nmin_try, 1.05]), k = np.exp(klogmin_try), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate, nw = nw, kw = kw, dw = model_parameters['dw']) * root_T1abs_n_single_layer_cryostat_echo(n = nmax_try, k = np.exp(klogmin_try), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate, nw = nw, kw = kw, dw = model_parameters['dw'])
                control_T1_kmax = root_T1abs_n_single_layer_cryostat_echo(n = max([nmin_try, 1.05]), k = np.exp(klogmax_try), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate, nw = nw, kw = kw, dw = model_parameters['dw']) * root_T1abs_n_single_layer_cryostat_echo(n = nmax_try, k = np.exp(klogmax_try), f = f_next, T0 = T0_next, T1 = T1_next, d0 = d_estimate, nw = nw, kw = kw, dw = model_parameters['dw'])
            else:
                raise ValueError('find_new_boundaries (cond_n): model %s is not available' % model)
        nmin = max([nmin_try, 1.05])
        nmax = nmax_try
        klogmin = klogmin_try
        klogmax = klogmax_try
        
        # check if while-loop stopped because of success or because interval became too small
        if control_T0_nmin >= 0 or control_T0_nmax >= 0 or control_T1_nmin >= 0 or control_T1_nmax >= 0:
            no_solution = True
        else:
            no_solution = False
        
    else:
        # At least one of the roots does not have a zero contour in the calculated interval. Bisection cannot be successful.
        n0_new = []
        n1_new = []
        k0_new = []
        k1_new = []
        n_intersect = None
        k_intersect = None
        nmin = min(n_lim)
        nmax = max(n_lim)
        klogmin = min(klog_lim)
        klogmax = max(klog_lim)
        no_solution = True  
    
    # uncomment the next line to see root map for debugging
    #plot_root_condition_map(n_axis = n_axis, klog_axis = klog_axis, root0 = root0, root1 = root1, n0 = n0, k0 = k0, n1 = n1, k1 = k1, n0_new = n0_new, k0_new = k0_new, n1_new = n1_new, k1_new = k1_new, n_intersect = n_intersect, k_intersect = k_intersect, nmin = nmin, nmax = nmax, klogmin = klogmin, klogmax = klogmax, title = '%f THz' % (f_next*1e-12))
    return [nmin, nmax], [klogmin, klogmax], no_solution

def interpolate_results_at_no_solution(f, n, k, d, no_solution, determine_thickness):
    """
    If no solution at a data point exists due to noise, use neighbouring points for linear interpolation. Else the original data is returned.
    """
    f_sol = []
    n_sol = []
    k_sol = []
    d_sol = []
    # find points where solution exists
    for i in range(len(f)):
        if no_solution[i] == False:
            f_sol.append(f[i])
            n_sol.append(n[i])
            k_sol.append(k[i])
            if determine_thickness == True:
                d_sol.append(d[i])
            else:
                d_sol.append(d)
    if len(f_sol) == len(f):
        # all solutions where found
        n_out = n
        k_out = k
        d_out = d
    elif len(f_sol) <= 1:
        n_out = n
        k_out = k
        d_out = d
        print('interpolate_results_at_no_solution: No solutions were found at any data point.')
    else:
        # linearly interpolate points with solutions to estimate values at points without solutions
        f_sol = np.array(f_sol)
        n_sol = np.array(n_sol)
        k_sol = np.array(k_sol)
        d_sol = np.array(d_sol)
        n_int = scipy.interpolate.interp1d(f_sol, n_sol, kind = 'linear', bounds_error = False, fill_value = (n_sol[0], n_sol[-1]))
        k_int = scipy.interpolate.interp1d(f_sol, k_sol, kind = 'linear', bounds_error = False, fill_value = (k_sol[0], k_sol[-1]))
        d_int = scipy.interpolate.interp1d(f_sol, d_sol, kind = 'linear', bounds_error = False, fill_value = (d_sol[0], d_sol[-1]))
        n_out = n_int(f)
        k_out = k_int(f)
        d_out = d_int(f)
    return n_out, k_out, d_out

def determine_fdep_optical_constants_in_interval(f_axis, T0_axis, T1_axis, d_estimate, n_lim, k_lim, model, model_parameters):
    # Frequency-dependent determination of n, k and sample thickness for various models by numerically solving the exact transmission equations
    # f_axis (np-array of floats): Frequency in THz
    # T0_axis (np-array of complex numbers): Transmission of main pulse
    # T1_axis (np-array of complex numbers): Transmission of 1st echo pulse
    # d_estimate (float): Estimated thickness of sample in m (needs to be estimated within one wavelength/n)
    # n_lim (list of 2 floats): Root finding limits for real part of refractive index
    # k_lim (list of 2 floats): Root finding limits for imaginary part of refractive index
    # model (string): Model name
    # model_parameters (dictionary): Model parameters, elements depend on model
    # returns: 
    #   n (np-array of floats): real part of refractive index 
    #   k (np-array of float): imaginary part of refractive index
    #   d (np-array of float): thickness of sample in m
    #   optimization (list of strings): "bisection" if bisection method was successful, "minimum" if contours do not cross in the investigated interval or "no_solution" if no appropriate interval was found at all.
    n = np.zeros(len(f_axis))
    k = np.zeros(len(f_axis))
    d = np.zeros(len(f_axis))
    optimization = [None] * len(f_axis)
    no_solution = [None] * len(f_axis)
    
    klog_lim = [np.log(min(k_lim)), np.log(max(k_lim))]
    n_lim0 = n_lim
    klog_lim0 = klog_lim
    for i in range(len(f_axis)):
        # estimate boundaries for root finding (avoid multiple sign changes)
        n_lim, klog_lim, no_solution[i] = find_new_boundaries(f_next = f_axis[i], T0_next = T0_axis[i], T1_next = T1_axis[i], d_estimate = d_estimate, n_lim = n_lim0, klog_lim = klog_lim0, model = model, model_parameters = model_parameters)
        # optimization at next data point
        if no_solution[i] == True:
            n[i] = 0.0
            k[i] = 0.0
            d[i] = 0.0
            optimization[i] = 'no_solution'
        else:
            n[i], k[i], d[i], optimization[i] = determine_optical_constants_in_interval(f = f_axis[i], T0 = T0_axis[i], T1 = T1_axis[i], d_estimate = d_estimate, n_lim = n_lim, klog_lim = klog_lim, model = model, model_parameters = model_parameters)
            d_estimate = d[i]
    # estimate data points without solution by linear interpolation from neighboring data points
    n, k, d = interpolate_results_at_no_solution(f = f_axis, n = n, k = k, d = d, no_solution = no_solution, determine_thickness = True)
    return n, k, d, optimization


def determine_optical_constants_in_interval_fixed_thickness(f, T0, d, x1_lim, x2_lim, model, model_parameters):
    # determine n, k and sample thickness for various models by numerically solving the exact transmission equations
    # f (float): Frequency in THz
    # T0 (complex): Transmission of main pulse
    # d (float): Thickness of sample in m
    # x1_lim (list of 2 floats): Root finding limits for auxilary variable x1 (x1 and x2 are chosen to rotate the contours in the n-k plane by 60° to avoid vertical contours)
    # x2_lim (list of 2 floats): Root finding limits for auxilary variable x2
    # model (string): Model name
    # model_parameters (dictionary): Model parameters, elements depend on model
    # returns: 
    #   n (float): real part of refractive index 
    #   k (float): imaginary part of refractive index
    #   optimization (string): "bisection" if bisection method was successful, "minimum" if contours do not cross in the investigated interval or "no_solution" if no appropriate interval was found at all.
    f = f * 1e12
    if model == 'single_layer':
        # check if there is a sign change between kmin and kmax
        cond_x2min = root_x2_single_layer_fixed_thickness(x2 = min(x2_lim), f = f, T0 = T0, d = d, x1min = min(x1_lim), x1max = max(x1_lim))
        cond_x2max = root_x2_single_layer_fixed_thickness(x2 = max(x2_lim), f = f, T0 = T0, d = d, x1min = min(x1_lim), x1max = max(x1_lim))
        if cond_x2min * cond_x2max <= 0:
            # sign change in k-interval: Apply bisection method (more precise)
            x2 = scipy.optimize.brentq(root_x2_single_layer_fixed_thickness, min(x2_lim), max(x2_lim), args = (f, T0, d, min(x1_lim), max(x1_lim)))
            optimization = 'bisection'
        else:
            # no sign change in k-interval: Apply minimization method (less precise)
            x2 = scipy.optimize.minimize_scalar(mini_x2_single_layer_fixed_thickness, bounds = (min(x2_lim), max(x2_lim)), args = (f, T0, d, min(x1_lim), max(x1_lim)), method = 'Bounded', options = {'xatol': 1e-16}).x
            optimization = 'minimum'
        # determine n for optimal k:
        x1_T0abs = scipy.optimize.brentq(root_T0abs_x1_single_layer_fixed_thickness, min(x1_lim), max(x1_lim), args = (x2, f, T0, d))
        x1_T0phase = scipy.optimize.brentq(root_T0phase_x1_single_layer_fixed_thickness, min(x1_lim), max(x1_lim), args = (x2, f, T0, d))
        x1 = (x1_T0abs + x1_T0phase) / 2
    elif model == 'single_layer_doublepass':
       # check if there is a sign change between kmin and kmax
       cond_x2min = root_x2_single_layer_doublepass_fixed_thickness(x2 = min(x2_lim), f = f, T0 = T0, d = d, x1min = min(x1_lim), x1max = max(x1_lim))
       cond_x2max = root_x2_single_layer_doublepass_fixed_thickness(x2 = max(x2_lim), f = f, T0 = T0, d = d, x1min = min(x1_lim), x1max = max(x1_lim))
       if cond_x2min * cond_x2max <= 0:
           # sign change in k-interval: Apply bisection method (more precise)
           x2 = scipy.optimize.brentq(root_x2_single_layer_doublepass_fixed_thickness, min(x2_lim), max(x2_lim), args = (f, T0, d, min(x1_lim), max(x1_lim)))
           optimization = 'bisection'
       else:
           # no sign change in k-interval: Apply minimization method (less precise)
           x2 = scipy.optimize.minimize_scalar(mini_x2_single_layer_doublepass_fixed_thickness, bounds = (min(x2_lim), max(x2_lim)), args = (f, T0, d, min(x1_lim), max(x1_lim)), method = 'Bounded', options = {'xatol': 1e-16}).x
           optimization = 'minimum'
       # determine n for optimal k:
       x1_T0abs = scipy.optimize.brentq(root_T0abs_x1_single_layer_doublepass_fixed_thickness, min(x1_lim), max(x1_lim), args = (x2, f, T0, d))
       x1_T0phase = scipy.optimize.brentq(root_T0phase_x1_single_layer_doublepass_fixed_thickness, min(x1_lim), max(x1_lim), args = (x2, f, T0, d))
       x1 = (x1_T0abs + x1_T0phase) / 2     
    else:
        raise ValueError('determine_optical_constants_in_interval: model %s is not available' % model)
    # calculate n and k from x1 and x2
    n = (x1 + np.sqrt(3) * x2)/2
    k = (-np.sqrt(3) * x1 + x2)/2
    return n, k, optimization

def plot_root_condition_map_fixed_thickness(X1, X2, root1, root2, x1_1, x2_1, x1_2, x2_2, x1_new, x2_1_new, x2_2_new, x1_intersect, x2_intersect, x1min, x1max, x2min, x2max, title):
    # plot root maps, contours, intersections and limits for debugging
    plt.clf()
    fig, ax = plt.subplots(1, 2, sharey= True)
    im = ax[0].pcolorfast(X1, X2, np.sign(root1[1:,1:]), cmap = 'bwr', vmin = -1, vmax = 1)
    ax[0].contour(X1, X2, root1, levels = [0], colors = 'k', linestyles = '-')
    ax[0].contour(X1, X2, root2, levels = [0], colors = 'k', linestyles = '--')
    if len(x2_1) != 0 and len(x1_1) != 0:
        ax[0].plot(x1_1, x2_1, 'y.')
        ax[0].plot(x1_new, x2_1_new, 'y-')
    if len(x2_2) != 0 and len(x1_2) != 0:
        ax[0].plot(x1_2, x2_2, 'g.')
        ax[0].plot(x1_new, x2_2_new, 'g-')
    if x1_intersect != None and x2_intersect != None:
        ax[0].plot(x1_intersect, x2_intersect, 'kx')
    rect = plt.Rectangle((x1min, x2min), x1max - x1min, x2max - x2min, facecolor = None, edgecolor = 'k', fill = False, linewidth = 0.5)
    ax[0].add_patch(rect)
    ax[0].set_title(title)
    im = ax[1].pcolorfast(X1, X2, np.sign(root2[1:,1:]), cmap = 'bwr', vmin = -1, vmax = 1)
    ax[1].contour(X1, X2, root1, levels = [0], colors = 'k', linestyles = '-')
    ax[1].contour(X1, X2, root2, levels = [0], colors = 'k', linestyles = '--')
    if len(x2_1) != 0 and len(x1_1) != 0:
        ax[1].plot(x1_1, x2_1, 'y.')
        ax[1].plot(x1_new, x2_1_new, 'y-')
    if len(x2_2) != 0 and len(x1_2) != 0:
        ax[1].plot(x1_2, x2_2, 'g.')
        ax[1].plot(x1_new, x2_2_new, 'g-')
    if x1_intersect != None and x2_intersect != None:
        ax[1].plot(x1_intersect, x2_intersect, 'kx')
    rect = plt.Rectangle((x1min, x2min), x1max - x1min, x2max - x2min, facecolor = None, edgecolor = 'k', fill = False, linewidth = 0.5)
    ax[1].add_patch(rect)
    plt.show()
    
    plt.clf()
    fig, ax = plt.subplots(1, 2, sharey= True)
    im = ax[0].pcolorfast(X1, X2, np.sign(root1[1:,1:]), cmap = 'bwr', vmin = -1, vmax = 1)
    ax[0].contour(X1, X2, root1, levels = [0], colors = 'k', linestyles = '-')
    ax[0].contour(X1, X2, root2, levels = [0], colors = 'k', linestyles = '--')
    if len(x2_1) != 0 and len(x1_1) != 0:
        ax[0].plot(x1_1, x2_1, 'y.')
        ax[0].plot(x1_new, x2_1_new, 'y-')
    if len(x2_2) != 0 and len(x1_2) != 0:
        ax[0].plot(x1_2, x2_2, 'g.')
        ax[0].plot(x1_new, x2_2_new, 'g-')
    if x1_intersect != None and x2_intersect != None:
        ax[0].plot(x1_intersect, x2_intersect, 'kx')
    rect = plt.Rectangle((x1min, x2min), x1max - x1min, x2max - x2min, facecolor = None, edgecolor = 'k', fill = False, linewidth = 0.5)
    ax[0].set_title(title)
    im = ax[1].pcolorfast(X1, X2, np.sign(root2[1:,1:]), cmap = 'bwr', vmin = -1, vmax = 1)
    ax[1].contour(X1, X2, root1, levels = [0], colors = 'k', linestyles = '-')
    ax[1].contour(X1, X2, root2, levels = [0], colors = 'k', linestyles = '--')
    if len(x2_1) != 0 and len(x1_1) != 0:
        ax[1].plot(x1_1, x2_1, 'y.')
        ax[1].plot(x1_new, x2_1_new, 'y-')
    if len(x2_2) != 0 and len(x1_2) != 0:
        ax[1].plot(x1_2, x2_2, 'g.')
        ax[1].plot(x1_new, x2_2_new, 'g-')
    if x1_intersect != None and x2_intersect != None:
        ax[1].plot(x1_intersect, x2_intersect, 'kx')
    rect = plt.Rectangle((x1min, x2min), x1max - x1min, x2max - x2min, facecolor = None, edgecolor = 'k', fill = False, linewidth = 0.5)
    ax[0].set_xlim(x1min, x1max)
    ax[0].set_ylim(x2min, x2max)
    ax[1].set_xlim(x1min, x1max)
    ax[1].set_ylim(x2min, x2max)
    plt.show()
    
    return

def find_new_boundaries_fixed_thickness(f_next, T0_next, d, n_prev, k_prev, x1_lim, x2_lim, model, model_parameters):
    # Finds neighbouring roots and poles to estimate limits of the parameter space for n and k at the next frequency
    # f_next (float): Frequency in THz (next data point)
    # T0_next (complex): Transmission of main pulse (next data point)
    # d (float): Thickness of sample in m
    # n_prev (float): Real part of refractive index of previous frequency data point
    # k_prev (float): Imaginray part of refractive index of previous frequency data point
    # x1_lim (list of two floats): Estimated limits of auxilary variable x1 for bisection
    # x2_lim (list of two floats): Estimated limits of auxilary variable x2 for bisection
    # model (string): Model name
    # model_parameters (dictionary): Model parameters, elements depend on model
    # returns:
    # x1_lim (list of 2 floats): New root finding limits for auxilary variable x1
    # x2_lim (list of 2 floats): New root finding limits for auxilary variable x2
    # no_solution (list of booleans): True if it is not possible to find bisection in both roots
    
    f_next = 1e12 * f_next
    c0 = 299792458
    # calculate n,k-landscapes for bisection condition
    if model == 'single_layer':
        nmin = n_prev - 0.9 * np.pi * c0 / (2 * np.pi * f_next * d)
        nmax = n_prev + 0.9 * np.pi * c0 / (2 * np.pi * f_next * d)
        kmin = max([np.exp(np.log(abs(k_prev)) - 0.9 * np.pi * c0 / (2 * np.pi * f_next * d)), 1e-12])#max([abs(k_prev) * 0.1, 1e-12])
        kmax = np.exp(np.log(abs(k_prev)) + 0.9 * np.pi * c0 / (2 * np.pi * f_next * d))#abs(k_prev) * 2#10
        # calculate auxilary variables
        x1_min = (nmin - np.sqrt(3) * kmax)/2
        x1_max = (nmax - np.sqrt(3) * kmin)/2
        x2_min = (np.sqrt(3) * nmin + kmin)/2
        x2_max = (np.sqrt(3) * nmax + kmax)/2
        x1_axis = np.linspace(x1_min, x1_max, 256)
        x2_axis = np.linspace(x2_min, x2_max, 256)
        X1, X2 = np.meshgrid(x1_axis, x2_axis)
        # calculate map to estimate contours
        root1 = root_T0abs_x1_single_layer_fixed_thickness(x1 = X1, x2 = X2, f = f_next, T0 = T0_next, d = d)
        root2 = root_T0phase_x1_single_layer_fixed_thickness(x1 = X1, x2 = X2, f = f_next, T0 = T0_next, d = d)
    elif model == 'single_layer_doublepass':
        nmin = n_prev - 1.5 * np.pi * c0 / (2 * np.pi * f_next * d * 2)
        nmax = n_prev + 1.5 * np.pi * c0 / (2 * np.pi * f_next * d * 2)
        kmin = max([abs(k_prev) * 0.1, 1e-12])
        kmax = abs(k_prev) * 10
        # calculate auxilary variables
        x1_min = (nmin - np.sqrt(3) * kmax)/2
        x1_max = (nmax - np.sqrt(3) * kmin)/2
        x2_min = (np.sqrt(3) * nmin + kmin)/2
        x2_max = (np.sqrt(3) * nmax + kmax)/2
        x1_axis = np.linspace(x1_min, x1_max, 256)
        x2_axis = np.linspace(x2_min, x2_max, 256)
        X1, X2 = np.meshgrid(x1_axis, x2_axis)
        # calculate map to estimate contours
        root1 = root_T0abs_x1_single_layer_doublepass_fixed_thickness(x1 = X1, x2 = X2, f = f_next, T0 = T0_next, d = d)
        root2 = root_T0phase_x1_single_layer_doublepass_fixed_thickness(x1 = X1, x2 = X2, f = f_next, T0 = T0_next, d = d)    
    else:
        raise ValueError('find_new_boundaries (cond_n): model %s is not available' % model)
    
    # find zero contour in root0 and root1
    grad1_x1 = np.diff(np.sign(root1), axis = 0)
    grad1_x2 = np.diff(np.sign(root1), axis = 1)
    x1_1 = X1[np.nonzero((grad1_x1[:,1:] + grad1_x2[1:,:]) < 0)]
    x2_1 = X2[np.nonzero((grad1_x1[:,1:] + grad1_x2[1:,:]) < 0)]
    inds = x1_1.argsort()
    x2_1 = x2_1[inds]
    x1_1 = x1_1[inds]
    x1_1, inds = np.unique(x1_1, return_index = True)
    x2_1 = x2_1[inds]
    
    grad2_x1 = np.diff(np.sign(root2), axis = 0)
    grad2_x2 = np.diff(np.sign(root2), axis = 1)
    x1_2 = X1[np.nonzero((grad2_x1[:,1:] + grad2_x2[1:,:]) > 0)]
    x2_2 = X2[np.nonzero((grad2_x1[:,1:] + grad2_x2[1:,:]) > 0)]
    inds = x1_2.argsort()
    x2_2 = x2_2[inds]
    x1_2 = x1_2[inds]
    x1_2, inds = np.unique(x1_2, return_index = True)
    x2_2 = x2_2[inds]
    if len(x1_1) > 1 and len(x1_2) > 1:
        # zero contours exist in both, root0 and root1.
        # estimate intersection point of contours via interpolation
        x1_new = x1_2
        x2_1_int = scipy.interpolate.interp1d(x = x1_1, y = x2_1, kind = 'linear', bounds_error = False, fill_value = 'extrapolate')
        x2_2_int = scipy.interpolate.interp1d(x = x1_2, y = x2_2, kind = 'linear', bounds_error = False, fill_value = 'extrapolate')
        x2_1_new = x2_1_int(x1_new)
        x2_2_new = x2_2_int(x1_new)
        x1_intersect = x1_new[np.argmin(abs(x2_1_new - x2_2_new))]
        x2_intersect = (x2_1_int(x1_intersect) + x2_2_int(x1_intersect))/2
           
        dx = 1e-3
        x1min_try = x1_min
        x1max_try = x1_max
        x2min_try = x2_min
        x2max_try = x2_max
        
        root1_x1min_x2min = 0
        root1_x1min_x2max = 0
        root1_x1max_x2min = 0
        root1_x1max_x2max = 0
        root2_x1min_x2min = 0
        root2_x1min_x2max = 0
        root2_x1max_x2min = 0
        root2_x1max_x2max = 0
        # make sure that root0 and root1 have the right sign change between x1min_try and x1max_try at x2min_try and x2max_try, respectively
        continue_search = True
        while continue_search == True and ((x1max_try - x1min_try) > 1e-6 and (x2max_try - x2min_try) > 1e-6): 
            if model == 'single_layer':
                root1_x1min_x2min = root_T0abs_x1_single_layer_fixed_thickness(x1 = x1min_try, x2 = x2min_try, f = f_next, T0 = T0_next, d = d)
                root1_x1max_x2min = root_T0abs_x1_single_layer_fixed_thickness(x1 = x1max_try, x2 = x2min_try, f = f_next, T0 = T0_next, d = d)
                root1_x1min_x2max = root_T0abs_x1_single_layer_fixed_thickness(x1 = x1min_try, x2 = x2max_try, f = f_next, T0 = T0_next, d = d)
                root1_x1max_x2max = root_T0abs_x1_single_layer_fixed_thickness(x1 = x1max_try, x2 = x2max_try, f = f_next, T0 = T0_next, d = d)
                root2_x1min_x2min = root_T0phase_x1_single_layer_fixed_thickness(x1 = x1min_try, x2 = x2min_try, f = f_next, T0 = T0_next, d = d)
                root2_x1max_x2min = root_T0phase_x1_single_layer_fixed_thickness(x1 = x1max_try, x2 = x2min_try, f = f_next, T0 = T0_next, d = d)
                root2_x1min_x2max = root_T0phase_x1_single_layer_fixed_thickness(x1 = x1min_try, x2 = x2max_try, f = f_next, T0 = T0_next, d = d)
                root2_x1max_x2max = root_T0phase_x1_single_layer_fixed_thickness(x1 = x1max_try, x2 = x2max_try, f = f_next, T0 = T0_next, d = d)
            elif model == 'single_layer_doublepass':
                x1min_try = x1min_try - dx
                x1max_try = x1max_try + dx
                x2min_try = x2min_try - dx/4
                x2max_try = x2max_try + dx/4
                root1_x1min_x2min = root_T0abs_x1_single_layer_doublepass_fixed_thickness(x1 = x1min_try, x2 = x2min_try, f = f_next, T0 = T0_next, d = d)
                root1_x1max_x2min = root_T0abs_x1_single_layer_doublepass_fixed_thickness(x1 = x1max_try, x2 = x2min_try, f = f_next, T0 = T0_next, d = d)
                root1_x1min_x2max = root_T0abs_x1_single_layer_doublepass_fixed_thickness(x1 = x1min_try, x2 = x2max_try, f = f_next, T0 = T0_next, d = d)
                root1_x1max_x2max = root_T0abs_x1_single_layer_doublepass_fixed_thickness(x1 = x1max_try, x2 = x2max_try, f = f_next, T0 = T0_next, d = d)
                root2_x1min_x2min = root_T0phase_x1_single_layer_doublepass_fixed_thickness(x1 = x1min_try, x2 = x2min_try, f = f_next, T0 = T0_next, d = d)
                root2_x1max_x2min = root_T0phase_x1_single_layer_doublepass_fixed_thickness(x1 = x1max_try, x2 = x2min_try, f = f_next, T0 = T0_next, d = d)
                root2_x1min_x2max = root_T0phase_x1_single_layer_doublepass_fixed_thickness(x1 = x1min_try, x2 = x2max_try, f = f_next, T0 = T0_next, d = d)
                root2_x1max_x2max = root_T0phase_x1_single_layer_doublepass_fixed_thickness(x1 = x1max_try, x2 = x2max_try, f = f_next, T0 = T0_next, d = d)
            else:
                raise ValueError('find_new_boundaries (cond_n): model %s is not available' % model)
            # determine limits for next try
            continue_search = False
            if root2_x1max_x2max <= 0:
                if root1_x1max_x2max > 0:
                    x1max_try = x1max_try - dx
                    x2max_try = x2max_try - dx
                else:
                    x1max_try = x1max_try + dx
                    x2max_try = x2max_try - dx
                continue_search = True
            if root2_x1min_x2min >= 0:
                if root1_x1min_x2min < 0:
                    x1min_try = x1min_try + dx
                    x2min_try = x2min_try + dx
                else:
                    x1min_try = x1min_try - dx
                    x2min_try = x2min_try + dx
                continue_search = True
            if root2_x1min_x2max >= 0:
                x2max_try = x2max_try - dx
                continue_search = True
            if root2_x1max_x2min <= 0:
                x2min_try = x2min_try + dx
                continue_search = True
        x1min = x1min_try
        x1max = x1max_try
        x2min = x2min_try
        x2max = x2max_try
        # check if while-loop stopped because of success or because interval became too small
        if root1_x1min_x2min >= 0 or root1_x1max_x2min <= 0 or root1_x1min_x2max >=0 or root1_x1max_x2max <= 0 or root2_x1min_x2min >= 0 or root2_x1max_x2min <= 0 or root2_x1min_x2max >= 0 or root2_x1max_x2max <= 0:
            no_solution = True
        else:
            no_solution = False        
    else:
        # At least one of the roots does not have a zero contour in the calculated interval. Bisection cannot be successful.
        x1_new = []
        x2_1_new = []
        x2_2_new = []
        x1_intersect = None
        x2_intersect = None
        x1min = min(x1_lim)
        x1max = max(x1_lim)
        x2min = min(x2_lim)
        x2max = max(x2_lim)
        no_solution = True  
    # uncomment the next line to see root map for debugging
    #plot_root_condition_map_fixed_thickness(X1 = X1, X2 = X2, root1 = root1, root2 = root2, x1_1 = x1_1, x2_1 = x2_1, x1_2 = x1_2, x2_2 = x2_2, x1_new = x1_new, x2_1_new = x2_1_new, x2_2_new = x2_2_new, x1_intersect = x1_intersect, x2_intersect = x2_intersect, x1min = x1min, x1max = x1max, x2min = x2min, x2max = x2max, title = '%f THz' % (f_next*1e-12))
    return [x1min, x1max], [x2min, x2max], no_solution

def determine_fdep_optical_constants_in_interval_fixed_thickness(f_axis, T0_axis, d, n_lim, k_lim, model, model_parameters):
    # Frequency-dependent determination of n, k and sample thickness for various models by numerically solving the exact transmission equations
    # f_axis (np-array of floats): Frequency in THz
    # T0_axis (np-array of complex numbers): Transmission of main pulse
    # d (float): Thickness of sample in m
    # n_lim (list of 2 floats): Root finding limits for real part of refractive index
    # k_lim (list of 2 floats): Root finding limits for imaginary part of refractive index
    # model (string): Model name
    # model_parameters (dictionary): Model parameters, elements depend on model
    # returns: 
    #   n (np-array of floats): real part of refractive index 
    #   k (np-array of float): imaginary part of refractive index
    #   optimization (list of strings): "bisection" if bisection method was successful, "minimum" if contours do not cross in the investigated interval or "no_solution" if no appropriate interval was found at all.
    n = np.zeros(len(f_axis))
    k = np.zeros(len(f_axis))
    optimization = [None] * len(f_axis)
    no_solution = [None] * len(f_axis)
    
    n_prev = (min(n_lim) + max(n_lim))/2
    k_prev = np.exp((np.log(min(k_lim)) + np.log(max(k_lim)))/2)
    x1_lim0 = [(min(n_lim) - max(k_lim))/np.sqrt(2), (max(n_lim) - min(k_lim))/np.sqrt(2)]
    x2_lim0 = [(min(n_lim) + min(k_lim))/np.sqrt(2), (max(n_lim) + max(k_lim))/np.sqrt(2)]
    for i in range(len(f_axis)):
        # estimate boundaries for root finding (avoid multiple sign changes)
        x1_lim, x2_lim, no_solution[i] = find_new_boundaries_fixed_thickness(f_next = f_axis[i], T0_next = T0_axis[i], d = d, n_prev = n_prev, k_prev = k_prev, x1_lim = x1_lim0, x2_lim = x2_lim0, model = model, model_parameters = model_parameters)
        # optimization at next data point
        if no_solution[i] == True:
            n[i] = 0.0
            k[i] = 0.0
            optimization[i] = 'no_solution'
        else:
            n[i], k[i], optimization[i] = determine_optical_constants_in_interval_fixed_thickness(f = f_axis[i], T0 = T0_axis[i], d = d, x1_lim = x1_lim, x2_lim = x2_lim, model = model, model_parameters = model_parameters)
            n_prev = n[i]
            k_prev = k[i]
    # estimate data points without solution by linear interpolation from neighboring data points
    n, k, d = interpolate_results_at_no_solution(f = f_axis, n = n, k = k, d = d, no_solution = no_solution, determine_thickness = False)
    return n, k, optimization

#%%
"""
Group delay method
"""

def determine_peak_time(t, S, t_interval, interpolation):
    if interpolation['interpolate'] == True:
        t_new = np.linspace(min(t_interval), max(t_interval), int((max(t_interval) - min(t_interval))/interpolation['dt']) + 1)
        S_int = scipy.interpolate.interp1d(x = t, y = S, kind = interpolation['method'], bounds_error = False, fill_value = 0.0)
        S_new = S_int(t_new)
    else:
        t_new = t[np.nonzero(np.logical_and(t >= min(t_interval), t <= max(t_interval)))]
        S_new = S[np.nonzero(np.logical_and(t >= min(t_interval), t <= max(t_interval)))]
    
    t_peak = t_new[np.argmax(abs(S_new))]
    return t_peak    
        
def determine_n_d_from_peak_delays(t, E_ref, E_sam, t_interval_ref, t_interval_sam0, t_interval_sam1, interpolation):
    # t: delay time in ps
    # E_ref: Reference (air) electric field
    # E_sam: Sample electrice field
    # t_interval_ref: 2-Element list that specifies the interval that contains the reference main pulse
    # t_interval_sam0: 2-Element list that specifies the interval that contains only the sample main pulse
    # t_interval_sam1: 2-Element list that specifies the interval that contains only the sample first echo pulse
    # interpolation: Dictionary with elements 'interpolate' (boolean), 'method' (string 'linear' or 'cubic'), 'dt' (float interpolated time sampling)
    # Output: Real refractive index n and sample thickness d in m
    
    # calculate n and d from delays between sample and ref main pulse and between sample main and first echo pulse
    # in this approach dispersion is neglected, i.e. n is frequency-independent
    # tp_sam_0 - tp_ref = t_10 = (n - 1) / c0 * d
    # tp_sam_1 - tp_sam_0 = t_21 = 2 * n / c0 * d
    
    tp_ref = 1e-12 * determine_peak_time(t = t, S = E_ref, t_interval = t_interval_ref, interpolation = interpolation)
    tp_sam0 = 1e-12 * determine_peak_time(t = t, S = E_sam, t_interval = t_interval_sam0, interpolation = interpolation)
    tp_sam1 = 1e-12 * determine_peak_time(t = t, S = E_sam, t_interval = t_interval_sam1, interpolation = interpolation)
    c0 = 299792458
    t10 = (tp_sam0 - tp_ref)
    t21 = (tp_sam1 - tp_sam0)
    n = 1 / (1 - 2 * t10/t21)
    d = c0 * (t21 / 2 - t10)
    return n, d

def determine_n_d_from_peak_delays_doublepass(t, E_ref, E_sam, t_interval_ref, t_interval_sam0, t_interval_sam1, interpolation):
    # t: delay time in ps
    # E_ref: Reference (air) electric field
    # E_sam: Sample electrice field
    # t_interval_ref: 2-Element list that specifies the interval that contains the reference main pulse
    # t_interval_sam0: 2-Element list that specifies the interval that contains only the sample main pulse
    # t_interval_sam1: 2-Element list that specifies the interval that contains only the sample first echo pulse
    # interpolation: Dictionary with elements 'interpolate' (boolean), 'method' (string 'linear' or 'cubic'), 'dt' (float interpolated time sampling)
    # Output: Real refractive index n and sample thickness d in m
    
    # calculate n and d from delays between sample and ref main pulse and between sample main and first echo pulse
    # in this approach dispersion is neglected, i.e. n is frequency-independent
    # tp_sam_0 - tp_ref = t_10 = (n - 1) / c0 * d
    # tp_sam_1 - tp_sam_0 = t_21 = 2 * n / c0 * d
    
    tp_ref = 1e-12 * determine_peak_time(t = t, S = E_ref, t_interval = t_interval_ref, interpolation = interpolation)
    tp_sam0 = 1e-12 * determine_peak_time(t = t, S = E_sam, t_interval = t_interval_sam0, interpolation = interpolation)
    tp_sam1 = 1e-12 * determine_peak_time(t = t, S = E_sam, t_interval = t_interval_sam1, interpolation = interpolation)
    c0 = 299792458
    t10 = (tp_sam0 - tp_ref)
    t21 = (tp_sam1 - tp_sam0)
    n = 1 / (1 - t10/t21)
    d = c0/2 * (t21 - t10)
    return n, d



#%%


def determine_windows(t, Et_ref, Et_sam, n_estimate, d_estimate, T_window, window_t_offset, window_type, window_parameter):
    # determine pulse maxima
    t0_ref = t[np.argmax(abs(Et_ref))] + window_t_offset
    t0_sam0 = t[np.argmax(abs(Et_sam))] + window_t_offset
    
    # estimate echo position
    t0_sam1 = t0_sam0 + 1e12 * 2 * n_estimate * d_estimate / 299792458
    
    # calculate three window functions
    if window_type == 'rect':
        window_ref = pulses.rectangular_window(t = t, t0 = t0_ref, T = T_window)
        window_sam0 = pulses.rectangular_window(t = t, t0 = t0_sam0, T = T_window)
        window_sam1 = pulses.rectangular_window(t = t, t0 = t0_sam1, T = T_window)    
    elif window_type == 'supergauss':
        window_ref = pulses.super_gauss(x = t, x0 = t0_ref, FWHM = T_window, n = window_parameter)
        window_sam0 = pulses.super_gauss(x = t, x0 = t0_sam0, FWHM = T_window, n = window_parameter)
        window_sam1 = pulses.super_gauss(x = t, x0 = t0_sam1, FWHM = T_window, n = window_parameter)
    elif window_type == 'Kaiser-Bessel':
        window_ref = pulses.Kaiser_Bessel_window(t = t - t0_ref, T = T_window, beta = window_parameter)
        window_sam0 = pulses.Kaiser_Bessel_window(t = t - t0_sam0, T = T_window, beta = window_parameter)
        window_sam1 = pulses.Kaiser_Bessel_window(t = t - t0_sam1, T = T_window, beta = window_parameter)
    else:
        raise ValueError('determine_windows: window_type %s is not available' % window_type)
    return window_ref, window_sam0, window_sam1, t0_ref, t0_sam0, t0_sam1

def calculate_windows(t, Et_ref, Et_sam0, Et_sam1, n_estimate, d_estimate, T_window, window_t_offset, window_type, window_parameter):
    # determine pulse maxima
    t0_ref = t[np.argmax(abs(Et_ref))] + window_t_offset
    t0_sam0 = t[np.argmax(abs(Et_sam0))] + window_t_offset
    if type(Et_sam1) == type(None):
        # estimate echo position
        t0_sam1 = t0_sam0 + 1e12 * 2 * n_estimate * d_estimate / 299792458
    else:
        t0_sam1 = t[np.argmax(abs(Et_sam1))] + window_t_offset
    
    # calculate three window functions
    if window_type == 'rect':
        window_ref = pulses.rectangular_window(t = t, t0 = t0_ref, T = T_window)
        window_sam0 = pulses.rectangular_window(t = t, t0 = t0_sam0, T = T_window)
        window_sam1 = pulses.rectangular_window(t = t, t0 = t0_sam1, T = T_window)    
    elif window_type == 'supergauss':
        window_ref = pulses.super_gauss(x = t, x0 = t0_ref, FWHM = T_window, n = window_parameter)
        window_sam0 = pulses.super_gauss(x = t, x0 = t0_sam0, FWHM = T_window, n = window_parameter)
        window_sam1 = pulses.super_gauss(x = t, x0 = t0_sam1, FWHM = T_window, n = window_parameter)
    elif window_type == 'Kaiser-Bessel':
        window_ref = pulses.Kaiser_Bessel_window(t = t - t0_ref, T = T_window, beta = window_parameter)
        window_sam0 = pulses.Kaiser_Bessel_window(t = t - t0_sam0, T = T_window, beta = window_parameter)
        window_sam1 = pulses.Kaiser_Bessel_window(t = t - t0_sam1, T = T_window, beta = window_parameter)
    else:
        raise ValueError('determine_windows: window_type %s is not available' % window_type)
    return window_ref, window_sam0, window_sam1, t0_ref, t0_sam0, t0_sam1

def determine_windows_and_groupddelayapproximation(t, Et_ref, Et_sam, options):
    # options is a dictionary that needs to contain the following entries:
    # model: string, that calls the right analysis model
    # n_estimate: float, estimated refractive index
    # d_estimate: float, estimated thickness
    # T_window: float, width of window
    # window_t_offset: float, shift of window with respect to pulse peaks
    # window_type: string, window-type specified in function "calculate_windows"
    # window_parameter: depends on window-type
    # interpolate: Boolean, if interpolation should be used in group delay approximation
    # interpolation_method: string, defined in scipy.optimize.interp1d
    # interpolation_dt_ratio: float
    
    # determine windows with estimated group delay between main and echo to isolate echo in first approximation
    window_ref, window_sam0, window_sam1, t0_ref, t0_sam0, t0_sam1 = calculate_windows(t = t, Et_ref = Et_ref, Et_sam0 = Et_sam, Et_sam1 = None, n_estimate = options['n_estimate'], d_estimate = options['d_estimate'], T_window = options['T_window'], window_t_offset = options['window_t_offset'], window_type = options['window_type'], window_parameter = options['window_parameter'])
    
    # refine window position with maximum in the interval determined in the previous step
    window_ref, window_sam0, window_sam1, t0_ref, t0_sam0, t0_sam1 = calculate_windows(t = t, Et_ref = Et_ref, Et_sam0 = Et_sam, Et_sam1 = Et_sam * window_sam1, n_estimate = options['n_estimate'], d_estimate = options['d_estimate'], T_window = options['T_window'], window_t_offset = options['window_t_offset'], window_type = options['window_type'], window_parameter = options['window_parameter'])
    
    # calculate n and d in group delay approximation by peak interpolation    
    if options['model'] == 'single_layer' or options['model'] == 'single_layer_cryostat_echo':
        n_peak, d_peak = determine_n_d_from_peak_delays(t = t, E_ref = Et_ref, E_sam = Et_sam, t_interval_ref = [t0_ref - options['T_window']/2, t0_ref + options['T_window']/2], t_interval_sam0 = [t0_sam0 - options['T_window']/2, t0_sam0 + options['T_window']/2], t_interval_sam1 = [t0_sam1 - options['T_window']/2, t0_sam1 + options['T_window']/2], interpolation = {'interpolate': options['interpolate'], 'method': options['interpolation_method'], 'dt': options['interpolation_dt_ratio'] * np.mean(np.diff(t))})
    elif options['model'] == 'single_layer_doublepass':
        n_peak, d_peak = determine_n_d_from_peak_delays_doublepass(t = t, E_ref = Et_ref, E_sam = Et_sam, t_interval_ref = [t0_ref - options['T_window']/2, t0_ref + options['T_window']/2], t_interval_sam0 = [t0_sam0 - options['T_window']/2, t0_sam0 + options['T_window']/2], t_interval_sam1 = [t0_sam1 - options['T_window']/2, t0_sam1 + options['T_window']/2], interpolation = {'interpolate': options['interpolate'], 'method': options['interpolation_method'], 'dt': options['interpolation_dt_ratio'] * np.mean(np.diff(t))})
    else:
        raise ValueError('determine_windows_and_groupdelayapproximation: model %s is not implemented' % options['model'])
    return window_ref, window_sam0, window_sam1, n_peak, d_peak
    
    

def plot_timedomain(sample_name, t, Et_ref, Et_sam, window_ref, window_sam0, window_sam1, outputpath, export):
    size = 128
    aspectratio = (1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))
    ax = fig.subplots()
    ax.plot(t, Et_ref / max(abs(Et_ref)), 'k-', label = 'air')
    ax.plot(t, Et_sam / max(abs(Et_ref)), 'r-', label = sample_name)
    ax.plot(t, window_ref, 'k--')
    ax.plot(t, window_sam0, 'r--')
    ax.plot(t, window_sam1, 'b--')
    ax.set_xlabel('$t$ [ps]')
    ax.set_ylabel('$E(t)$ [arb. u.]')
    ax.legend(loc = 'upper right')
    ax.grid()
    ax.set_xlim(min(t), max(t))
    if export == True:
        plt.savefig(r'%s\%s_timedomain.png' % (outputpath, sample_name), dpi = 600, bbox_inches = 'tight')
    plt.show()
    return

def plot_frequencydomain(sample_name, f, Ef_ref, Ef_sam0, Ef_sam1, outputpath, export):
    size = 128
    aspectratio = (1 + np.sqrt(5))/4
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))
    ax = fig.subplots(2, 1, sharex = True)
    ax[0].plot(f, abs(Ef_ref)/max(abs(Ef_ref)), 'k-', label = 'air')
    ax[0].plot(f, abs(Ef_sam0)/max(abs(Ef_ref)), 'r-', label = '%s, main' % sample_name)
    ax[0].plot(f, abs(Ef_sam1)/max(abs(Ef_ref)), 'b-', label = '%s, echo' % sample_name)
    ax[1].set_xlabel('$f$ [THz]')
    ax[0].set_ylabel('$|E(f)|$ [arb. u.]')
    ax[0].legend(loc = 'upper right')
    ax[0].grid()
    ax[0].set_xlim(min(f), max(f))
    ax[1].plot(f, np.gradient(np.unwrap(np.angle(Ef_ref)))/np.gradient(f), 'k--', label = 'air')
    ax[1].plot(f, np.gradient(np.unwrap(np.angle(Ef_sam0)))/np.gradient(f), 'r--', label = '%s, main' % sample_name)
    ax[1].plot(f, np.gradient(np.unwrap(np.angle(Ef_sam1)))/np.gradient(f), 'b--', label = '%s, echo' % sample_name)
    ax[1].set_ylabel('arg$(E(f))$ [rad]')
    ax[1].set_xlim(min(f), max(f)) 
    ax[1].grid()
    if export == True:
        plt.savefig(r'%s\%s_frequencydomain.png' % (outputpath, sample_name), dpi = 600, bbox_inches = 'tight')
    plt.show()
    
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))
    ax = fig.subplots(2, 1, sharex = True)
    ax[0].plot(f, np.log(abs(Ef_ref)/max(abs(Ef_ref)))/np.log(10), 'k-', label = 'air')
    ax[0].plot(f, np.log(abs(Ef_sam0)/max(abs(Ef_ref)))/np.log(10), 'r-', label = '%s, main' % sample_name)
    ax[0].plot(f, np.log(abs(Ef_sam1)/max(abs(Ef_ref)))/np.log(10), 'b-', label = '%s, echo' % sample_name)
    ax[1].set_xlabel('$f$ [THz]')
    ax[0].set_ylabel('$|E(f)|$ [arb. u.]')
    ax[0].legend(loc = 'upper right')
    ax[0].grid()
    ax[0].set_xlim(min(f), max(f))
    ax[1].plot(f, np.gradient(np.unwrap(np.angle(Ef_ref)))/np.gradient(f), 'k--', label = 'air')
    ax[1].plot(f, np.gradient(np.unwrap(np.angle(Ef_sam0)))/np.gradient(f), 'r--', label = '%s, main' % sample_name)
    ax[1].plot(f, np.gradient(np.unwrap(np.angle(Ef_sam1)))/np.gradient(f), 'b--', label = '%s, echo' % sample_name)
    ax[1].set_ylabel('arg$(E(f))$ [rad]')
    ax[1].set_xlim(min(f), max(f))
    ax[1].grid()
    if export == True:
        plt.savefig(r'%s\%s_frequencydomain2.png' % (outputpath, sample_name), dpi = 600, bbox_inches = 'tight')
    plt.show()
    
    return


def plot_optical_constants(f, n, k, d, Ef_ref, Ef_sam0, Ef_sam1, n2, k2, d2, sample_name, outputpath, export):
    size = 150 # mm
    aspectratio = 0.7#(1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))     
    ax = fig.subplots(nrows = 4, sharex = True)
    ax[0].plot(f, n, 'k-')
    ax[0].plot(f, n2, 'r-')
    ax[0].set_ylabel('$n$')
    ax[0].grid()
    ax[1].plot(f, k, 'k-')
    ax[1].plot(f, k2, 'r-')
    ax[1].set_ylabel('$k$')
    ax[1].grid()
    ax[2].plot(f, d * 1e3, 'k-')
    ax[2].plot(f, d2 * 1e3, 'r-')
    ax[2].set_ylabel('$d$ [mm]')
    ax[2].grid()
    ax[3].plot(f, np.log(abs(Ef_ref)/max(abs(Ef_ref)))/np.log(10), 'k-', label = 'air')
    ax[3].plot(f, np.log(abs(Ef_sam0)/max(abs(Ef_ref)))/np.log(10), 'r-', label = 'sample, main')
    ax[3].plot(f, np.log(abs(Ef_sam1)/max(abs(Ef_ref)))/np.log(10), 'b-', label = 'sample, echo')
    ax[3].set_xlabel('$f$ [THz]')
    ax[3].set_ylabel('log$_{10}(|E(f)|)$ [norm.]')
    ax[3].legend(loc = 'lower left')
    ax[3].grid()
    ax[3].set_xlim(min(f), max(f))
    if export == True:
        plt.savefig(r'%s\%s_opticalconstants.png' % (outputpath, sample_name), dpi = 600, bbox_inches = 'tight')
    plt.show()
    return


def plot_optical_constants_varibel_thickness(f, n, k, d, Ef_ref, Ef_sam0, Ef_sam1, sample_name, outputpath, export):
    size = 150 # mm
    aspectratio = 0.7#(1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))     
    ax = fig.subplots(nrows = 4, sharex = True)
    ax[0].plot(f, n, 'k-')
    ax[0].set_ylabel('$n$')
    ax[0].grid()
    ax[1].plot(f, k, 'k-')
    ax[1].set_ylabel('$k$')
    ax[1].grid()
    ax[2].plot(f, d * 1e3, 'k-')
    ax[2].set_ylabel('$d$ [mm]')
    ax[2].grid()
    ax[3].plot(f, np.log(abs(Ef_ref)/max(abs(Ef_ref)))/np.log(10), 'k-', label = 'air')
    ax[3].plot(f, np.log(abs(Ef_sam0)/max(abs(Ef_ref)))/np.log(10), 'r-', label = 'sample, main')
    ax[3].plot(f, np.log(abs(Ef_sam1)/max(abs(Ef_ref)))/np.log(10), 'b-', label = 'sample, echo')
    ax[3].set_xlabel('$f$ [THz]')
    ax[3].set_ylabel('log$_{10}(|E(f)|)$ [norm.]')
    ax[3].legend(loc = 'lower left')
    ax[3].grid()
    ax[3].set_xlim(min(f), max(f))
    if export == True:
        plt.savefig(r'%s\%s_opticalconstants.png' % (outputpath, sample_name), dpi = 600, bbox_inches = 'tight')
    plt.show()
    return

def plot_optical_constants_fixed_thickness(f, Ef_ref, Ef_sam0, Ef_sam1, n2, k2, d2, sample_name, outputpath, export):
    size = 150 # mm
    aspectratio = 0.7#(1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))     
    ax = fig.subplots(nrows = 4, sharex = True)
    ax[0].plot(f, n2, 'r-')
    ax[0].set_ylabel('$n$')
    ax[0].grid()
    ax[1].plot(f, k2, 'r-')
    ax[1].set_ylabel('$k$')
    ax[1].grid()
    ax[2].plot(f, d2 * 1e3, 'r-')
    ax[2].set_ylabel('$d$ [mm]')
    ax[2].grid()
    ax[3].plot(f, np.log(abs(Ef_ref)/max(abs(Ef_ref)))/np.log(10), 'k-', label = 'air')
    ax[3].plot(f, np.log(abs(Ef_sam0)/max(abs(Ef_ref)))/np.log(10), 'r-', label = 'sample, main')
    ax[3].plot(f, np.log(abs(Ef_sam1)/max(abs(Ef_ref)))/np.log(10), 'b-', label = 'sample, echo')
    ax[3].set_xlabel('$f$ [THz]')
    ax[3].set_ylabel('log$_{10}(|E(f)|)$ [norm.]')
    ax[3].legend(loc = 'lower left')
    ax[3].grid()
    ax[3].set_xlim(min(f), max(f))
    if export == True:
        plt.savefig(r'%s\%s_opticalconstants_fixed_thickness.png' % (outputpath, sample_name), dpi = 600, bbox_inches = 'tight')
    plt.show()
    return


def export_optical_constants(dataset, f, Ef_ref, Ef_sam0, Ef_sam1, n, k, d, optimization, n_peak, d_peak):
    file = open(r'%s\%s_optimization_result.txt' % (dataset['outputpath'], dataset['name']), 'w')
    file.write('# %s' % time.strftime("%d.%m.%Y %H:%M:%S"))
    for option in list(dataset):
        file.write('\n# %s = %s' % (option, str(dataset[option])))
    file.write('\nn_peak = %.10f' % n_peak)
    file.write('\nd_peak = %.10f' % d_peak)
    file.write('\n# %7s %8s %8s %10s %10s %10s %11s %11s %11s %11s' % ('f [THz]', 'n', 'k', 'd [m]', 'Re(Ef_ref)', 'Im(Ef_ref)', 'Re(Ef_sam0)', 'Im(Ef_sam0)', 'Re(Ef_sam1)', 'Im(Ef_sam1)'))
    for i in range(len(f)):
        file.write('\n%9.4f %8.6f %8.6f %10.8f %10.6f %10.6f %11.6f %11.6f %11.6f %11.6f %s' % (f[i], n[i], k[i], d[i], Ef_ref[i].real, Ef_ref[i].imag, Ef_sam0[i].real, Ef_sam0[i].imag, Ef_sam1[i].real, Ef_sam1[i].imag, optimization[i]))
    file.close()
    return    

def export_optical_constants2(dataset, f, n, k, d, d_err, optimization):
    file = open(r'%s\%s_optimization_result_fixedthickness.txt' % (dataset['outputpath'], dataset['name']), 'w')
    file.write('# %s' % time.strftime("%d.%m.%Y %H:%M:%S"))
    for option in list(dataset):
        file.write('\n# %s = %s' % (option, str(dataset[option])))
    file.write('\nd_eff [m] = %.12f' % d)
    file.write('\nd_eff_err [m] = %.12f' % d_err)
    file.write('\n# %7s %8s %8s' % ('f [THz]', 'n', 'k'))
    for i in range(len(f)):
        file.write('\n%9.4f %8.6f %8.6f %s' % (f[i], n[i], k[i], optimization[i]))
    file.close()
    return

def export_TDS_data(t, Et_ref, Et_sam, dataset):
    file = open(r'%s\%s_TDS_rawdata.txt' % (dataset['outputpath'], dataset['name']), 'w')
    file.write('# Original sample file: %s' % dataset['file_sample'])
    file.write('\n# Original reference file: %s' % dataset['file_reference'])
    file.write('\n%11s, %13s, %13s' % ('Time_abs/ps', 'Signal_sample', 'Signal_ref'))
    for i in range(len(t)):
        file.write('\n%11.3f, %13.6f, %13.6f' % (t[i], Et_sam[i], Et_ref[i]))
    file.close()
    return



def evaluate_Teraflash_dataset_weighted(dataset, export):
    if export == True:
        # prepare export path
        if not os.path.exists(dataset['outputpath']):
            os.makedirs(dataset['outputpath'])  
        #shutil.copy(r'C:\Users\nbeermann\Python Skripte\AG_pulses_m8.py', r'%s\%s' % (dataset['outputpath'], 'AG_pulses_m8.py'))
        #shutil.copy(r'C:\Users\nbeermann\Python Skripte\AG_THzTDS_m10.py', r'%s\%s' % (dataset['outputpath'], 'AG_THzTDS_m10.py'))
        #shutil.copy(__file__, r'%s\%s_%s' % (dataset['outputpath'], time.strftime("%d%m%Y_%H%M%S"), 'executed_script.py'))
    
    # read data
    if dataset['file_reference'] == None:
        t, Et_sam, Et_ref = read_Teraflash_1file(file = dataset['file_sample'])
    else:
        t, Et_sam, Et_ref = read_Teraflash_2files(ref_file = dataset['file_reference'], sam_file = dataset['file_sample'])
    if export == True:
        # copy raw data to export path
        export_TDS_data(t = t, Et_ref = Et_ref, Et_sam = Et_sam, dataset = dataset)
    
    if dataset['subtract_background'] == True:
        t, Et_BG, Et_BG2 = read_Teraflash_2files(ref_file = dataset['file_background'], sam_file = dataset['file_background']) # same file is read twice to avoid programming new read-function
        Et_ref = Et_ref - Et_BG
        Et_sam = Et_sam - Et_BG
    
    # determine windows
    window_ref, window_sam0, window_sam1, n_peak, d_peak = determine_windows_and_groupddelayapproximation(t = t, Et_ref = Et_ref, Et_sam = Et_sam, options = dataset)
    plot_timedomain(sample_name = dataset['name'], t = t, Et_ref = Et_ref, Et_sam = Et_sam, window_ref = window_ref, window_sam0 = window_sam0, window_sam1 = window_sam1, outputpath = dataset['outputpath'], export = export)
    
    # FFT of data times window
    w, Ew_ref = pulses.t2w_transform(t = t, Et = Et_ref * window_ref)
    w, Ew_sam0 = pulses.t2w_transform(t = t, Et = Et_sam * window_sam0)
    w, Ew_sam1 = pulses.t2w_transform(t = t, Et = Et_sam * window_sam1)
    
    # truncate frequency axis
    Ew_ref = Ew_ref[np.nonzero(np.logical_and(w >= dataset['fmin'] * 2 * np.pi, w <= dataset['fmax'] * 2 * np.pi))]
    Ew_sam0 = Ew_sam0[np.nonzero(np.logical_and(w >= dataset['fmin'] * 2 * np.pi, w <= dataset['fmax'] * 2 * np.pi))]
    Ew_sam1 = Ew_sam1[np.nonzero(np.logical_and(w >= dataset['fmin'] * 2 * np.pi, w <= dataset['fmax'] * 2 * np.pi))]
    w = w[np.nonzero(np.logical_and(w >= dataset['fmin'] * 2 * np.pi, w <= dataset['fmax'] * 2 * np.pi))]
    f = w / (2*np.pi)
    plot_frequencydomain(sample_name = dataset['name'], f = f, Ef_ref = Ew_ref, Ef_sam0 = Ew_sam0, Ef_sam1 = Ew_sam1, outputpath = dataset['outputpath'], export = export)
    
    # determine n, k, d
    n, k, d, opti = determine_fdep_optical_constants_in_interval(f_axis = f, T0_axis = np.nan_to_num(Ew_sam0 / Ew_ref), T1_axis = np.nan_to_num(Ew_sam1 / Ew_ref), d_estimate = d_peak, n_lim = [dataset['nmin'], dataset['nmax']], k_lim = [dataset['kmin'], dataset['kmax']], model = dataset['model'], model_parameters = dataset['model_parameters'])
    if export == True:
        export_optical_constants(dataset = dataset, f = f, Ef_ref = Ew_ref, Ef_sam0 = Ew_sam0, Ef_sam1 = Ew_sam1, n = n, k = k, d = d, optimization = opti, n_peak = n_peak, d_peak = d_peak) 
   
    plot_optical_constants_varibel_thickness(f = f, n = n, k = k, d = d,  Ef_ref = Ew_ref, Ef_sam0 = Ew_sam0, Ef_sam1 = Ew_sam1, sample_name = dataset['name'], outputpath = dataset['outputpath'], export = export)
    return {'t': t, 'Et_ref': Et_ref, 'Et_sam': Et_sam, 'f': w/(2*np.pi), 'Ef_ref': Ew_ref, 'W_ref': window_ref, 'W_sam0': window_sam0, 'W_sam1': window_sam1, 'Ef_sam0': Ew_sam0, 'Ef_sam1': Ew_sam1, 'n': n, 'k': k, 'd': d, 'optimization': opti}
 

def evaluate_Teraflash_dataset_weighted_fixed_thickness(dataset, export, weighted, weighted_error):
    if export == True:
        # prepare export path
        if not os.path.exists(dataset['outputpath']):
            os.makedirs(dataset['outputpath'])  
        #shutil.copy(r'C:\Users\nbeermann\Python Skripte\AG_pulses_m8.py', r'%s\%s' % (dataset['outputpath'], 'AG_pulses_m8.py'))
        #shutil.copy(r'C:\Users\nbeermann\Python Skripte\AG_THzTDS_m10.py', r'%s\%s' % (dataset['outputpath'], 'AG_THzTDS_m10.py'))
        #shutil.copy(__file__, r'%s\%s_%s' % (dataset['outputpath'], time.strftime("%d%m%Y_%H%M%S"), 'executed_script.py'))
    
    # read data
    if dataset['file_reference'] == None:
        t, Et_sam, Et_ref = read_Teraflash_1file(file = dataset['file_sample'])
    else:
        t, Et_sam, Et_ref = read_Teraflash_2files(ref_file = dataset['file_reference'], sam_file = dataset['file_sample'])
    if export == True:
        # copy raw data to export path
        export_TDS_data(t = t, Et_ref = Et_ref, Et_sam = Et_sam, dataset = dataset)
    
    if dataset['subtract_background'] == True:
        t, Et_BG, Et_BG2 = read_Teraflash_2files(ref_file = dataset['file_background'], sam_file = dataset['file_background']) # same file is read twice to avoid programming new read-function
        Et_ref = Et_ref - Et_BG
        Et_sam = Et_sam - Et_BG
    
    # determine windows
    window_ref, window_sam0, window_sam1, n_peak, d_peak = determine_windows_and_groupddelayapproximation(t = t, Et_ref = Et_ref, Et_sam = Et_sam, options = dataset)
    plot_timedomain(sample_name = dataset['name'], t = t, Et_ref = Et_ref, Et_sam = Et_sam, window_ref = window_ref, window_sam0 = window_sam0, window_sam1 = window_sam1, outputpath = dataset['outputpath'], export = export)
    
    # FFT of data times window
    w, Ew_ref = pulses.t2w_transform(t = t, Et = Et_ref * window_ref)
    w, Ew_sam0 = pulses.t2w_transform(t = t, Et = Et_sam * window_sam0)
    w, Ew_sam1 = pulses.t2w_transform(t = t, Et = Et_sam * window_sam1)
    
    # truncate frequency axis
    Ew_ref = Ew_ref[np.nonzero(np.logical_and(w >= dataset['fmin'] * 2 * np.pi, w <= dataset['fmax'] * 2 * np.pi))]
    Ew_sam0 = Ew_sam0[np.nonzero(np.logical_and(w >= dataset['fmin'] * 2 * np.pi, w <= dataset['fmax'] * 2 * np.pi))]
    Ew_sam1 = Ew_sam1[np.nonzero(np.logical_and(w >= dataset['fmin'] * 2 * np.pi, w <= dataset['fmax'] * 2 * np.pi))]
    w = w[np.nonzero(np.logical_and(w >= dataset['fmin'] * 2 * np.pi, w <= dataset['fmax'] * 2 * np.pi))]
    f = w / (2*np.pi)
    
    d_eff = weighted
    d_std = weighted_error
    
    
    # determine n, k, d with average thickness
    n2, k2, opti2 = determine_fdep_optical_constants_in_interval_fixed_thickness(f_axis = f, T0_axis = np.nan_to_num(Ew_sam0 / Ew_ref), d = d_eff, n_lim = [n_peak - 1, n_peak + 1], k_lim = [dataset['kmin'], dataset['kmax']], model = dataset['model'], model_parameters = dataset['model_parameters'])
    d2 = np.ones(len(f)) * d_eff
    if export == True:
        export_optical_constants2(dataset = dataset, f = f, n = n2, k = k2, d = d_eff, d_err = d_std, optimization = opti2)
    
    plot_optical_constants_fixed_thickness(f = f, n2 = n2, k2 = k2, d2 = d2, Ef_ref = Ew_ref, Ef_sam0 = Ew_sam0, Ef_sam1 = Ew_sam1, sample_name = dataset['name'], outputpath = dataset['outputpath'], export = export)
    return {'t': t, 'Et_ref': Et_ref, 'Et_sam': Et_sam, 'f': w/(2*np.pi), 'Ef_ref': Ew_ref, 'W_ref': window_ref, 'W_sam0': window_sam0, 'W_sam1': window_sam1, 'Ef_sam0': Ew_sam0, 'Ef_sam1': Ew_sam1,  'n2': n2, 'k2': k2, 'd2': d2, 'optimization2': opti2}
 
    
    
def read_mean_data(basepath, material, T):
    file = open(r'%s\%s_%.0fK_nkd_mean.txt' % (basepath, material, T))
    lines = file.readlines()
    file.close()
    d = 0
    d_err = 0
    f = []
    n = []
    n_err = []
    k = []
    k_err = []
    for line in lines:
        if line.startswith('#'):
            continue
        elif line.startswith('d '):
            s = line.split()
            d = float(s[-1])
            continue
        elif line.startswith('d_err'):
            s = line.split()
            d_err = float(s[-1])
            continue
        s = line.split()
        f.append(float(s[0]))
        n.append(float(s[1]))
        n_err.append(float(s[2]))
        k.append(float(s[3]))
        k_err.append(float(s[4]))
    data = {'T': T, 'f': np.array(f), 'n': np.array(n), 'n_err': np.array(n_err), 'k': np.array(k), 'k_err': np.array(k_err), 'd': d, 'd_err': d_err}
    return data 


def export_average_nkd_optimazation_result(file_paths, base_path):

    # Load data for all files
    all_data = []
    for file_path in file_paths:
        data = load_data_first(file_path)
        all_data.append(data)

    # Initialize arrays to store component-wise averages and standard deviations
    num_files = len(all_data)
    num_entries = len(all_data[0]['n'])  # Assuming all files have the same length for 'n'

    opti = np.array(all_data[0]['solution'])
    
    freq = np.array(all_data[0]['f [THz]'])

    # Initialize arrays for average and standard deviation
    avg_n_variabel = np.zeros(num_entries)
    avg_k_variabel = np.zeros(num_entries)
    avg_d_variabel = np.zeros(num_entries)
    std_n_variabel = np.zeros(num_entries)
    std_k_variabel = np.zeros(num_entries)
    std_d_variabel = np.zeros(num_entries)


    for data in all_data:
        avg_n_variabel += np.array(data['n']) / num_files
        avg_k_variabel += np.array(data['k']) / num_files
        avg_d_variabel += np.array(data['d [m]']) / num_files 
        #all_data_in_one += data['d [m]'] 
    

    # Calculate standard deviations component-wise using np.std
    for i in range(num_entries):
        std_n_variabel[i] = np.std([data['n'][i] for data in all_data])
        std_k_variabel[i] = np.std([data['k'][i] for data in all_data])
        std_d_variabel[i] = np.std([data['d [m]'][i] for data in all_data])
    
    
    # Convert thickness and error from meters to micrometers
    avg_d_variabel *= 1e6  # Convert average thickness from meters to micrometers
    std_d_variabel *= 1e6  # Convert standard deviation of thickness from meters to micrometers
    
      
    #lw_plot = 1

    size = 150 # mm
    aspectratio = 0.7#(1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))     
    ax = fig.subplots(nrows = 3, sharex = True)
    ax[0].plot(freq, avg_n_variabel, 'r-')
    ax[0].fill_between(x = freq, y1 = avg_n_variabel - std_n_variabel, y2 = avg_n_variabel + std_n_variabel, color = 'k', alpha = 0.5)
    ax[0].set_ylabel('$n$')
    ax[0].grid()
    #ax[0].set_ylim([np.round(avg_n_variabel[10],2)-0.5, np.round(avg_n_variabel[10],2)+1])
    ax[1].plot(freq, avg_k_variabel, 'k-')
    ax[1].fill_between(x = freq, y1 = avg_k_variabel - std_k_variabel, y2 = avg_k_variabel + std_k_variabel, color = 'k', alpha = 0.5)
    ax[1].set_ylabel('$k$')
    ax[1].grid()
    #ax[1].set_ylim([np.round(avg_k_variabel[10],2)-0.005, np.round(avg_k_variabel[10],2)+ 0.05])
    ax[2].plot(freq, avg_d_variabel, 'b-')
    ax[2].fill_between(x = freq, y1 = avg_d_variabel - std_d_variabel, y2 = avg_d_variabel + std_d_variabel, color = 'k', alpha = 0.5)
    ax[2].set_ylabel('$d$ [mum]')
    ax[2].grid()
    #ax[2].set_ylim([np.round(avg_d_variabel[30],2)-200, np.round(avg_d_variabel[30],2)+ 50])
    plt.savefig(r'%s\opticalconstants_variabel.png' % (base_path), dpi = 600, bbox_inches = 'tight')
    plt.show()
    

    d_raw = {'freq':freq, 'avg_n_variabel':avg_n_variabel,'std_n_variabel':std_n_variabel,  'avg_k_variabel':avg_k_variabel,'std_k_variabel':std_k_variabel, 'avg_d_variabel':avg_d_variabel,'std_d_variabel':std_d_variabel, 'solution': opti }
    df_raw = pd.DataFrame(data=d_raw)
    df_all = pd.concat([df_raw], axis=1)
    df_all.to_excel(base_path + "nkd_mean_optimization.xlsx")
     
    return



def load_data_first(filename):
    headers = ['f [THz]', 'n', 'k', 'd [m]', 'Re(Ef_ref)', 'Im(Ef_ref)', 
               'Re(Ef_sam0)', 'Im(Ef_sam0)', 'Re(Ef_sam1)', 'Im(Ef_sam1)', 'solution']
    data = {header: [] for header in headers}

    data_started = False

    with open(filename, 'r') as file:
        for line in file:
            line = line.strip()

            # Skip lines until we find the header line
            if not data_started:
                if line.startswith('# f [THz]'):
                    data_started = True
                continue

            # Process data lines
            if data_started and line and not line.startswith('#'):
                values = line.split()
                if len(values) == len(headers):
                    for header, value in zip(headers, values):
                        try:
                            data[header].append(float(value))
                        except ValueError:
                            data[header].append(value)  # Handle non-numeric values

    return data



def replace_zeros_with_value(arr):
    """
    Replace all elements equal to 0 in a NumPy array with 0.01.
    
    :param arr: NumPy array of numbers
    :return: Modified NumPy array with zeros replaced by 0.01
    """
    arr[arr == 0] = 0.0001
    return arr


def calculate_weighted_thickness(file_paths, f_min , f_max, number_measurements):
    # Load data for all files
    all_data = []
    for file_path in file_paths:
        data = load_data_first(file_path)
        all_data.append(data)

    # Initialize arrays to store component-wise averages and standard deviations
    num_files = len(all_data)
    num_entries = len(all_data[0]['n'])  # Assuming all files have the same length for 'n'
    
    opti = np.array(all_data[0]['solution'])

    freq = np.array(all_data[0]['f [THz]'])

    # Initialize arrays for average and standard deviation
    avg_d = np.zeros(num_entries)
    std_d = np.zeros(num_entries)


    for data in all_data:
        avg_d += np.array(data['d [m]']) / num_files 
    

    # Calculate standard deviations component-wise using np.std
    for i in range(num_entries):
        std_d[i] = np.std([data['d [m]'][i] for data in all_data])
    
    
    # Convert thickness and error from meters to micrometers
    avg_d *= 1e6  # Convert average thickness from meters to micrometers
    std_d *= 1e6  # Convert standard deviation of thickness from meters to micrometers



    # Find indices where frequency is within the specified range and opti[i] is 'bisection'
    indices = np.where((freq >= f_min) & (freq <= f_max) & (opti == 'bisection'))[0]
    
    

    avg_d_range  = np.zeros(len(indices), dtype=np.float64)
    std_d_range  = np.zeros(len(indices), dtype=np.float64)
    std_d_range1 = np.zeros(len(indices), dtype=np.float64)
    
       

    avg_d_range   = np.array(avg_d[indices], dtype=np.float64)
    std_d_range1  = np.array(std_d[indices], dtype=np.float64)
    
    
    print(std_d_range1)
    std_d_range   = replace_zeros_with_value(std_d_range1)

    weighted_average = np.sum(avg_d_range/(std_d_range**2))/(np.sum( 1/(std_d_range**2)))
    
    print(weighted_average)
    

    N = len(avg_d_range)

    weighted_average_error = 1/(np.sqrt(number_measurements)) *   np.sqrt(   N/(N-1) *(   (   np.sum(avg_d_range**2/(std_d_range**2)))     /(np.sum(1/(std_d_range**2)))       - weighted_average**2   )) 


    plt.title('Variante 5: ' + np.str(round(weighted_average, 4)) + ' mum +- ' + np.str(round(weighted_average_error, 4)) + ' mum'  )
    plt.errorbar(freq, avg_d,std_d)
    plt.axvline(x = f_min, color = 'k', ls = '--')
    plt.axvline(x = f_max, color = 'k', ls = '--')
    plt.axhline(y=weighted_average, color='r', linestyle='-')
    plt.legend()
    plt.xlabel('Freq [THz]')
    plt.ylabel('Thickness [mum]')
    ax=plt.gca()
    ax.set_xlim([0.1,3.0])
    ax.set_ylim([weighted_average-300,weighted_average+300])
    plt.savefig('rawdata', bbox_inches='tight')
    plt.show()
    plt.clf()
    
    
    all_in_one = []   
    indices2 = np.where((freq >= f_min) & (freq <= f_max) & (opti == 'bisection'))[0]
    

        
        
    all_in_one = []

    for i in range(0, num_files):
        a = np.average(np.array(all_data[i]['d [m]'])[indices2])
        #b = (np.array(all_data[i]['d [m]'])[indices])
        all_in_one.append(a*1e6)
           
        
    print(all_in_one)
    # Calculate average thickness in the specified range
    statistic_average =  np.mean(all_in_one) 
    statistic_average_error = np.std(all_in_one) 
    print(statistic_average_error)   
    
    return weighted_average*1e-6, weighted_average_error*1e-6, statistic_average*1e-6, statistic_average_error*1e-6



#%%
"""
averaging
"""

def read_TDS_eval_files(file):
    data = open(file)
    lines = data.readlines()
    data.close()
    f = []
    n = []
    k = []
    d = []
    Ef_ref = []
    Ef_sam0 = []
    Ef_sam1 = []
    n_peak = 0
    d_peak = 0
    optimization = []
    for line in lines:
        if line.startswith('#'):
            continue
        elif line.startswith('n_peak'):
            s = line.split()
            n_peak = float(s[-1])
            continue
        elif line.startswith('d_peak'):
            s = line.split()
            d_peak = float(s[-1])
            continue
        s = line.split()
        f.append(float(s[0]))
        n.append(float(s[1]))
        k.append(float(s[2]))
        d.append(float(s[3]))
        Ef_ref.append(complex(float(s[4]), float(s[5])))
        Ef_sam0.append(complex(float(s[6]), float(s[7])))
        Ef_sam1.append(complex(float(s[8]), float(s[9])))
        optimization.append(s[10])
    S = {'f': np.array(f),
         'n': np.array(n),
         'k': np.array(k),
         'd': np.array(d),
         'Ef_ref': np.array(Ef_ref),
         'Ef_sam0': np.array(Ef_sam0),
         'Ef_sam1': np.array(Ef_sam1),
         'n_peak': n_peak,
         'd_peak': d_peak,
         'optimization': optimization}
    return S

def read_TDS_eval_files_fixedthickness(file):
    data = open(file)
    lines = data.readlines()
    data.close()
    f = []
    n = []
    k = []
    d = 0
    d_err = 0
    optimization = []
    for line in lines:
        if line.startswith('#'):
            continue
        elif line.startswith('d_eff [m]'):
            s = line.split()
            d = float(s[-1])
            continue
        elif line.startswith('d_eff_err'):
            s = line.split()
            d_err = float(s[-1])
            continue
        s = line.split()
        f.append(float(s[0]))
        n.append(float(s[1]))
        k.append(float(s[2]))
        optimization.append(s[3])
    S = {'f': np.array(f),
         'n': np.array(n),
         'k': np.array(k),
         'd': d,
         'd_err': d_err,
         'optimization': optimization}
    return S

def get_average(S_list):
    n_mean = np.zeros(len(S_list[0]['f']))
    k_mean = np.zeros(len(S_list[0]['f']))
    d_mean = np.zeros(len(S_list[0]['f']))
    n_std = np.zeros(len(S_list[0]['f']))
    k_std = np.zeros(len(S_list[0]['f']))
    d_std = np.zeros(len(S_list[0]['f']))
    neff_mean = 0
    neff_std = 0
    deff_mean = 0
    deff_std = 0
    Ef_ref_mean = np.zeros(len(S_list[0]['f']), dtype = complex)
    Ef_sam0_mean = np.zeros(len(S_list[0]['f']), dtype = complex)
    Ef_sam1_mean = np.zeros(len(S_list[0]['f']), dtype = complex)
    Ef_ref_std = np.zeros(len(S_list[0]['f']), dtype = complex)
    Ef_sam0_std = np.zeros(len(S_list[0]['f']), dtype = complex)
    Ef_sam1_std = np.zeros(len(S_list[0]['f']), dtype = complex)
    for S in S_list:
        n_mean += S['n']
        k_mean += S['k']
        d_mean += S['d']
        neff_mean += np.sum(S['n'] * abs(S['Ef_ref'])) / np.sum(abs(S['Ef_ref']))
        deff_mean += np.sum(S['d'] * abs(S['Ef_ref'])) / np.sum(abs(S['Ef_ref']))
        Ef_ref_mean += S['Ef_ref'] / max(abs(S['Ef_ref']))
        Ef_sam0_mean += S['Ef_sam0'] / max(abs(S['Ef_ref']))
        Ef_sam1_mean += S['Ef_sam1'] / max(abs(S['Ef_ref']))
    n_mean = n_mean / len(S_list)
    k_mean = k_mean / len(S_list)
    d_mean = d_mean / len(S_list)
    neff_mean = neff_mean / len(S_list)
    deff_mean = deff_mean / len(S_list)
    Ef_ref_mean = Ef_ref_mean / len(S_list)
    Ef_sam0_mean = Ef_sam0_mean / len(S_list)
    Ef_sam1_mean = Ef_sam1_mean / len(S_list)
    for S in S_list:
        n_std += (S['n'] - n_mean)**2
        k_std += (S['k'] - k_mean)**2
        d_std += (S['d'] - d_mean)**2
        neff_std += (np.sum(S['n'] * abs(S['Ef_ref'])) / np.sum(abs(S['Ef_ref'])) - neff_mean)**2
        deff_std += (np.sum(S['d'] * abs(S['Ef_ref'])) / np.sum(abs(S['Ef_ref'])) - deff_mean)**2
        Ef_ref_std += (S['Ef_ref'] / max(abs(S['Ef_ref'])) - Ef_ref_mean)**2
        Ef_sam0_std += (S['Ef_sam0'] / max(abs(S['Ef_ref'])) - Ef_sam0_mean)**2
        Ef_sam1_std += (S['Ef_sam1'] / max(abs(S['Ef_ref'])) - Ef_sam1_mean)**2
    n_std = np.sqrt(1/(len(S_list)-1) * n_std)
    k_std = np.sqrt(1/(len(S_list)-1) * k_std)
    d_std = np.sqrt(1/(len(S_list)-1) * d_std)
    neff_std = np.sqrt(1/(len(S_list)-1) * neff_std)
    deff_std = np.sqrt(1/(len(S_list)-1) * deff_std)
    Ef_ref_std = np.sqrt(1/(len(S_list)-1) * Ef_ref_std)
    Ef_sam0_std = np.sqrt(1/(len(S_list)-1) * Ef_sam0_std)
    Ef_sam1_std = np.sqrt(1/(len(S_list)-1) * Ef_sam1_std)
    n_peak_mean = 0
    d_peak_mean = 0
    n_peak_std = 0
    d_peak_std = 0
    for S in S_list:
        n_peak_mean += S['n_peak']
        d_peak_mean += S['d_peak']
    n_peak_mean = n_peak_mean / len(S_list)
    d_peak_mean = d_peak_mean / len(S_list)
    for S in S_list:
        n_peak_std += (S['n_peak'] - n_peak_mean)**2
        d_peak_std += (S['d_peak'] - d_peak_mean)**2
    n_peak_std = np.sqrt(1/(len(S_list)-1) * n_peak_std)
    d_peak_std = np.sqrt(1/(len(S_list)-1) * d_peak_std)
        
    mean_values = {'n': n_mean,
                   'n_err': n_std,
                   'n_eff': neff_mean,
                   'n_eff_err': neff_std,
                   'n_peak': n_peak_mean,
                   'n_peak_err': n_peak_std,
                   'k': k_mean,
                   'k_err': k_std,
                   'd': d_mean,
                   'd_err': d_std,
                   'd_eff': deff_mean,
                   'd_eff_err': deff_std,
                   'd_peak': d_peak_mean,
                   'd_peak_err': d_peak_std,
                   'Ef_ref': Ef_ref_mean,
                   'Ef_ref_err': Ef_ref_std,
                   'Ef_sam0': Ef_sam0_mean,
                   'Ef_sam0_err': Ef_sam0_std,
                   'Ef_sam1': Ef_sam1_mean,
                   'Ef_sam1_err': Ef_sam1_std}
    return mean_values

def get_average_fixed_thickness(S_list):
    n_mean = np.zeros(len(S_list[0]['f']))    
    k_mean = np.zeros(len(S_list[0]['f']))
    d_mean_oben = 0.0
    d_mean_unten = 0.0
    n_std = np.zeros(len(S_list[0]['f']))    
    k_std = np.zeros(len(S_list[0]['f']))
    d_std_oben = 0.0
    d_std_unten = 0.0
    for S in S_list:
        n_mean += S['n']
        k_mean += S['k']
        d_mean_oben += S['d'] / S['d_err']**2
        d_mean_unten += 1 / S['d_err']**2
    n_mean = n_mean / len(S_list)
    k_mean = k_mean / len(S_list)
    d_mean = d_mean_oben / d_mean_unten
    for S in S_list:
        n_std += (S['n'] - n_mean)**2
        k_std += (S['k'] - k_mean)**2
        d_std_oben += S['d']**2 / S['d_err']**2
        d_std_unten += 1 / S['d_err']**2
    n_std = np.sqrt(1 / (len(S_list) - 1) * n_std)
    k_std = np.sqrt(1 / (len(S_list) - 1) * k_std)
    d_std = np.sqrt(len(S)/(len(S)-1) * (d_std_oben / d_std_unten - d_mean**2))
    mean_values = {'n': n_mean, 'n_err': n_std, 'k': k_mean, 'k_err': k_std, 'd': d_mean, 'd_err': d_std}
    return mean_values

def get_average_fixed_thickness_weighted(S_list, weighted, weighted_error):
    n_mean = np.zeros(len(S_list[0]['f']))    
    k_mean = np.zeros(len(S_list[0]['f']))
    d_mean_oben = 0.0
    d_mean_unten = 0.0
    n_std = np.zeros(len(S_list[0]['f']))    
    k_std = np.zeros(len(S_list[0]['f']))
    d_std_oben = 0.0
    d_std_unten = 0.0
    for S in S_list:
        n_mean += S['n']
        k_mean += S['k']
        d_mean_oben += S['d'] / S['d_err']**2
        d_mean_unten += 1 / S['d_err']**2
    n_mean = n_mean / len(S_list)
    k_mean = k_mean / len(S_list)
    d_mean = d_mean_oben / d_mean_unten
    d_mean_weighted = weighted
    for S in S_list:
        n_std += (S['n'] - n_mean)**2
        k_std += (S['k'] - k_mean)**2
        d_std_oben += S['d']**2 / S['d_err']**2
        d_std_unten += 1 / S['d_err']**2
    n_std = np.sqrt(1 / (len(S_list) - 1) * n_std)
    k_std = np.sqrt(1 / (len(S_list) - 1) * k_std)
    d_std = np.sqrt(len(S)/(len(S)-1) * (d_std_oben / d_std_unten - d_mean**2))
    d_std_weighted = weighted_error
    mean_values = {'n': n_mean, 'n_err': n_std, 'k': k_mean, 'k_err': k_std, 'd': d_mean, 'd_err': d_std,  'd_weighted': d_mean_weighted, 'd_err_weighted': d_std_weighted}
    return mean_values




def export_average_data(S_list, outputfile, header):
    mean = get_average_fixed_thickness(S_list)
    file = open(outputfile, 'w')
    if header == '':
        file.write('#')
    else:
        file.write(header)
    file.write('\nd [mm] = %.6f' % (mean['d']*1e3))
    file.write('\nd_err [mm] = %.6f' %( mean['d_err']*1e3))
    file.write('\n# %7s %10s %10s %10s %10s' % ('f [THz]', 'n', 'n_err', 'k', 'k_err'))
    for i in range(len(S_list[0]['f'])):
        file.write('\n%9.4f %10.6f %10.6f %10.6f %10.6f' % (S_list[0]['f'][i], mean['n'][i], mean['n_err'][i], mean['k'][i], mean['k_err'][i]))
    file.close()
    return


def export_average_data_weighted(S_list, weighted, weighted_error, outputfile, header):
    mean = get_average_fixed_thickness(S_list)
    file = open(outputfile, 'w')
    if header == '':
        file.write('#')
    else:
        file.write(header)
    file.write('\nd [mm] = %.6f' % (mean['d']*1e3))
    file.write('\nd_err [mm] = %.6f' %( mean['d_err']*1e3))
    file.write('\nd_weighted [mm] = %.6f' %( weighted*1e3))
    file.write('\nd_err_weighted [mm] = %.6f' %( weighted_error*1e3))
    file.write('\n# %7s %10s %10s %10s %10s' % ('f [THz]', 'n', 'n_err', 'k', 'k_err'))
    for i in range(len(S_list[0]['f'])):
        file.write('\n%9.4f %10.6f %10.6f %10.6f %10.6f' % (S_list[0]['f'][i], mean['n'][i], mean['n_err'][i], mean['k'][i], mean['k_err'][i]))
    file.close()
    return



def read_and_export_TDS_data_weighted_Temp(basepath, material, export, header, measurements, weighted,  weighted_error, temp):
    S1 = []
    S2 = []
    for i in measurements:
        S1.append(read_TDS_eval_files(file = r'%s\%s_%iK_%i_optimization_result.txt' % (basepath, material, temp, i)))
        S2.append(read_TDS_eval_files_fixedthickness(file = r'%s\%s_%iK_%i_optimization_result_fixedthickness.txt' % (basepath, material, temp, i)))
    mean1 = get_average(S1)
    mean2 = get_average_fixed_thickness_weighted(S2, weighted,  weighted_error)
    if export == True:
        export_average_data_weighted(S_list = S2, weighted = weighted, weighted_error = weighted_error, outputfile = r'%s\%s_%iK_nkd_mean.txt' % (basepath, material, temp), header = header)
    return S1[1]['f'], mean1, mean2


        
def read_and_export_TDS_data(basepath, material, export, header, measurements):
    S1 = []
    S2 = []
    for i in measurements:
        S1.append(read_TDS_eval_files(file = r'%s\%s_RT_%i_optimization_result.txt' % (basepath, material, i)))
        S2.append(read_TDS_eval_files_fixedthickness(file = r'%s\%s_RT_%i_optimization_result_fixedthickness.txt' % (basepath, material, i)))
    mean1 = get_average(S1)
    mean2 = get_average_fixed_thickness(S2)
    if export == True:
        export_average_data(S_list = S2, outputfile = r'%s\%s_RT_nkd_mean.txt' % (basepath, material), header = header)
    return S1[1]['f'], mean1, mean2


def read_and_export_TDS_data_weighted(basepath, material, export, header, measurements, weighted,  weighted_error):
    S1 = []
    S2 = []
    for i in measurements:
        S1.append(read_TDS_eval_files(file = r'%s\%s_RT_%i_optimization_result.txt' % (basepath, material, i)))
        S2.append(read_TDS_eval_files_fixedthickness(file = r'%s\%s_RT_%i_optimization_result_fixedthickness.txt' % (basepath, material, i)))
    mean1 = get_average(S1)
    mean2 = get_average_fixed_thickness_weighted(S2, weighted,  weighted_error)
    if export == True:
        export_average_data_weighted(S_list = S2, weighted = weighted, weighted_error = weighted_error, outputfile = r'%s\%s_RT_nkd_mean.txt' % (basepath, material), header = header)
    return S1[1]['f'], mean1, mean2

def read_and_export_TDS_data_weighted2(basepath, material, export, header, measurements, weighted,  weighted_error):
    S1 = []
    S2 = []
    for i in measurements:
        S1.append(read_TDS_eval_files(file = r'%s\%s_RT_%i_optimization_result.txt' % (basepath, material, i+1)))
        S2.append(read_TDS_eval_files_fixedthickness(file = r'%s\%s_RT_%i_optimization_result_fixedthickness.txt' % (basepath, material, i+1)))
    mean1 = get_average(S1)
    mean2 = get_average_fixed_thickness_weighted(S2, weighted,  weighted_error)
    if export == True:
        export_average_data_weighted(S_list = S2, weighted = weighted, weighted_error = weighted_error, outputfile = r'%s\%s_RT_nkd_mean.txt' % (basepath, material), header = header)
    return S1[1]['f'], mean1, mean2

def plot_nkd(f, mean, mean_fixed_thickness, flim, nlim, klim, dlim, f_interval, title, export, path):
    lw_plot = 1

    size = 150 # mm
    aspectratio = 0.7#(1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))     
    ax = fig.subplots(nrows = 4, sharex = True)
    ax[0].fill_between(x = f, y1 = mean['n'] - mean['n_err'], y2 = mean['n'] + mean['n_err'], color = 'k', alpha = 0.5)
    ax[0].plot(f, mean['n'], 'k-', label = 'variable thickness')
    ax[0].fill_between(x = f, y1 = mean_fixed_thickness['n'] - mean_fixed_thickness['n_err'], y2 = mean_fixed_thickness['n'] + mean_fixed_thickness['n_err'], color = 'r', alpha = 0.5)
    ax[0].plot(f, mean_fixed_thickness['n'], 'r-', label = 'fixed thickness')
    ax[0].set_ylim(min(nlim), max(nlim))
    ax[0].set_ylabel('$n$')
    ax[0].grid()
    ax[0].legend(loc = 'upper left')
    ax[0].set_title(title)
    ax[1].fill_between(x = f, y1 = mean['k'] - mean['k_err'], y2 = mean['k'] + mean['k_err'], color = 'k', alpha = 0.5)
    ax[1].plot(f, mean['k'], 'k-')
    ax[1].fill_between(x = f, y1 = mean_fixed_thickness['k'] - mean_fixed_thickness['k_err'], y2 = mean_fixed_thickness['k'] + mean_fixed_thickness['k_err'], color = 'r', alpha = 0.5)
    ax[1].plot(f, mean_fixed_thickness['k'], 'r-')
    ax[1].set_ylabel('$k$')
    ax[1].set_ylim(min(klim), max(klim))
    ax[1].grid()
    ax[2].fill_between(x = f, y1 = 1e6*(mean['d'] - mean['d_err']), y2 = 1e6*(mean['d'] + mean['d_err']), color = 'k', alpha = 0.5)
    ax[2].plot(f, 1e6*mean['d'], 'k-')
    ax[2].fill_between(x = f, y1 = 1e6*np.ones(len(f))*(mean_fixed_thickness['d'] - mean_fixed_thickness['d_err']), y2 = 1e6*(mean_fixed_thickness['d'] + mean_fixed_thickness['d_err']), color = 'r', alpha = 0.5)
    ax[2].plot(f, 1e6*np.ones(len(f))*mean_fixed_thickness['d'], 'r-')
    if len(f_interval) == 2:
        ax[2].vlines(x = min(f_interval), ymin = min(dlim), ymax = max(dlim), color = 'k', linestyle = 'dashed')
        ax[2].vlines(x = max(f_interval), ymin = min(dlim), ymax = max(dlim), color = 'k', linestyle = 'dashed')
    ax[2].set_ylabel('$d$ [$\mu$m]')
    ax[2].grid()
    ax[2].text(x = max(f_interval) + 0.5, y = min(dlim) + 0.25 * (max(dlim) - min(dlim)), s = r'$%.2f +/- %.2f \mu$m' % (1e6*mean_fixed_thickness['d'], 1e6*mean_fixed_thickness['d_err']), color = 'r')
    ax[2].set_ylim(min(dlim), max(dlim))
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_ref'] - mean['Ef_ref_err']), y2 = abs(mean['Ef_ref'] + mean['Ef_ref_err']), color = 'k', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_ref']), 'k-', label = 'ref')
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_sam0'] - mean['Ef_sam0_err']), y2 = abs(mean['Ef_sam0'] + mean['Ef_sam0_err']), color = 'r', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_sam0']), 'r-', label = 'main')
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_sam1'] - mean['Ef_sam1_err']), y2 = abs(mean['Ef_sam1'] + mean['Ef_sam1_err']), color = 'b', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_sam1']), 'b-', label = 'echo')
    if len(f_interval) == 2:
        ax[3].vlines(x = min(f_interval), ymin = 1e-5, ymax = 1.5, color = 'k', linestyle = 'dashed')
        ax[3].vlines(x = max(f_interval), ymin = 1e-5, ymax = 1.5, color = 'k', linestyle = 'dashed')
        ax[3].annotate('thickness interval', xy = (max(f_interval), 3e-2), xytext = (max(f_interval) + 0.5, 3e-5), arrowprops = dict(arrowstyle = '->', linewidth = lw_plot, color = 'k'), color = 'k')
    ax[3].set_yscale('log', nonposy = 'clip')
    ax[3].set_xlabel('$f$ [THz]')
    ax[3].set_ylabel('$|E(f)|$ [norm.]')
    ax[3].legend(loc = 'lower left')
    ax[3].grid()
    ax[3].set_xlim(min(flim), max(flim))
    ax[3].set_ylim(1e-5, 1.5)
    if export == True:
        plt.savefig(r'%s\nkd.png' % path, dpi = 600, bbox_inches = 'tight')
    plt.show()
    return


def plot_nkd_weighted(f, mean, mean_fixed_thickness, flim, nlim, klim, dlim, f_interval, title, export, path):
    lw_plot = 1

    size = 150 # mm
    aspectratio = 0.7#(1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))     
    ax = fig.subplots(nrows = 4, sharex = True)
    ax[0].fill_between(x = f, y1 = mean['n'] - mean['n_err'], y2 = mean['n'] + mean['n_err'], color = 'k', alpha = 0.5)
    ax[0].plot(f, mean['n'], 'k-', label = 'variable thickness')
    ax[0].fill_between(x = f, y1 = mean_fixed_thickness['n'] - mean_fixed_thickness['n_err'], y2 = mean_fixed_thickness['n'] + mean_fixed_thickness['n_err'], color = 'r', alpha = 0.5)
    ax[0].plot(f, mean_fixed_thickness['n'], 'r-', label = 'fixed thickness')
    ax[0].set_ylim(min(nlim), max(nlim))
    ax[0].set_ylabel('$n$')
    ax[0].grid()
    ax[0].legend(loc = 'upper left')
    ax[0].set_title(title)
    ax[1].fill_between(x = f, y1 = mean['k'] - mean['k_err'], y2 = mean['k'] + mean['k_err'], color = 'k', alpha = 0.5)
    ax[1].plot(f, mean['k'], 'k-')
    ax[1].fill_between(x = f, y1 = mean_fixed_thickness['k'] - mean_fixed_thickness['k_err'], y2 = mean_fixed_thickness['k'] + mean_fixed_thickness['k_err'], color = 'r', alpha = 0.5)
    ax[1].plot(f, mean_fixed_thickness['k'], 'r-')
    ax[1].set_ylabel('$k$')
    ax[1].set_ylim(min(klim), max(klim))
    ax[1].grid()
    ax[2].fill_between(x = f, y1 = 1e6*(mean['d'] - mean['d_err']), y2 = 1e6*(mean['d'] + mean['d_err']), color = 'k', alpha = 0.5)
    ax[2].plot(f, 1e6*mean['d'], 'k-')
    ax[2].fill_between(x = f, y1 = 1e6*np.ones(len(f))*(mean_fixed_thickness['d_weighted'] - mean_fixed_thickness['d_err_weighted']), y2 = 1e6*(mean_fixed_thickness['d_weighted'] + mean_fixed_thickness['d_err_weighted']), color = 'r', alpha = 0.5)
    ax[2].plot(f, 1e6*np.ones(len(f))*mean_fixed_thickness['d_weighted'], 'r-')
    if len(f_interval) == 2:
        ax[2].vlines(x = min(f_interval), ymin = min(dlim), ymax = max(dlim), color = 'k', linestyle = 'dashed')
        ax[2].vlines(x = max(f_interval), ymin = min(dlim), ymax = max(dlim), color = 'k', linestyle = 'dashed')
    ax[2].set_ylabel('$d$ [$\mu$m]')
    ax[2].grid()
    ax[2].text(x = max(f_interval) + 0.5, y = min(dlim) + 0.25 * (max(dlim) - min(dlim)), s = r'$%.2f +/- %.2f \mu$m' % (1e6*mean_fixed_thickness['d_weighted'], 1e6*mean_fixed_thickness['d_err_weighted']), color = 'r')
    ax[2].text(x = max(f_interval) + 0.5, y = min(dlim) + 0.05 * (max(dlim) - min(dlim)), s = r'$%.2f +/- %.2f \mu$m' % (1e6*mean_fixed_thickness['d'], 1e6*mean_fixed_thickness['d_err']), color = 'b')
    ax[2].set_ylim(min(dlim), max(dlim))
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_ref'] - mean['Ef_ref_err']), y2 = abs(mean['Ef_ref'] + mean['Ef_ref_err']), color = 'k', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_ref']), 'k-', label = 'ref')
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_sam0'] - mean['Ef_sam0_err']), y2 = abs(mean['Ef_sam0'] + mean['Ef_sam0_err']), color = 'r', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_sam0']), 'r-', label = 'main')
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_sam1'] - mean['Ef_sam1_err']), y2 = abs(mean['Ef_sam1'] + mean['Ef_sam1_err']), color = 'b', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_sam1']), 'b-', label = 'echo')
    if len(f_interval) == 2:
        ax[3].vlines(x = min(f_interval), ymin = 1e-5, ymax = 1.5, color = 'k', linestyle = 'dashed')
        ax[3].vlines(x = max(f_interval), ymin = 1e-5, ymax = 1.5, color = 'k', linestyle = 'dashed')
        ax[3].annotate('thickness interval', xy = (max(f_interval), 3e-2), xytext = (max(f_interval) + 0.5, 3e-5), arrowprops = dict(arrowstyle = '->', linewidth = lw_plot, color = 'k'), color = 'k')
    ax[3].set_yscale('log', nonpositive='clip')
    ax[3].set_xlabel('$f$ [THz]')
    ax[3].set_ylabel('$|E(f)|$ [norm.]')
    ax[3].legend(loc = 'lower left')
    ax[3].grid()
    ax[3].set_xlim(min(flim), max(flim))
    ax[3].set_ylim(1e-5, 1.5)
    if export == True:
        plt.savefig(r'%s\nkd.png' % path, dpi = 600, bbox_inches = 'tight')
    plt.show()
    return

def plot_nkd_weighted2(f, mean, mean_fixed_thickness, statistic_average, statistic_average_error, flim, nlim, klim, dlim, f_interval, title, export, path):
    lw_plot = 1

    size = 150 # mm
    aspectratio = 0.7#(1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))     
    ax = fig.subplots(nrows = 4, sharex = True)
    ax[0].fill_between(x = f, y1 = mean['n'] - mean['n_err'], y2 = mean['n'] + mean['n_err'], color = 'k', alpha = 0.5)
    ax[0].plot(f, mean['n'], 'k-', label = 'variable thickness')
    ax[0].fill_between(x = f, y1 = mean_fixed_thickness['n'] - mean_fixed_thickness['n_err'], y2 = mean_fixed_thickness['n'] + mean_fixed_thickness['n_err'], color = 'r', alpha = 0.5)
    ax[0].plot(f, mean_fixed_thickness['n'], 'r-', label = 'fixed thickness')
    ax[0].set_ylim(min(nlim), max(nlim))
    ax[0].set_ylabel('$n$')
    ax[0].grid()
    ax[0].legend(loc = 'upper left')
    ax[0].set_title(title)
    ax[1].fill_between(x = f, y1 = mean['k'] - mean['k_err'], y2 = mean['k'] + mean['k_err'], color = 'k', alpha = 0.5)
    ax[1].plot(f, mean['k'], 'k-')
    ax[1].fill_between(x = f, y1 = mean_fixed_thickness['k'] - mean_fixed_thickness['k_err'], y2 = mean_fixed_thickness['k'] + mean_fixed_thickness['k_err'], color = 'r', alpha = 0.5)
    ax[1].plot(f, mean_fixed_thickness['k'], 'r-')
    ax[1].set_ylabel('$k$')
    ax[1].set_ylim(min(klim), max(klim))
    ax[1].grid()
    ax[2].fill_between(x = f, y1 = 1e6*(mean['d'] - mean['d_err']), y2 = 1e6*(mean['d'] + mean['d_err']), color = 'k', alpha = 0.5)
    ax[2].plot(f, 1e6*mean['d'], 'k-')
    ax[2].fill_between(x = f, y1 = 1e6*np.ones(len(f))*(mean_fixed_thickness['d_weighted'] - mean_fixed_thickness['d_err_weighted']), y2 = 1e6*(mean_fixed_thickness['d_weighted'] + mean_fixed_thickness['d_err_weighted']), color = 'r', alpha = 0.5)
    ax[2].plot(f, 1e6*np.ones(len(f))*mean_fixed_thickness['d_weighted'], 'r-')
    if len(f_interval) == 2:
        ax[2].vlines(x = min(f_interval), ymin = min(dlim), ymax = max(dlim), color = 'k', linestyle = 'dashed')
        ax[2].vlines(x = max(f_interval), ymin = min(dlim), ymax = max(dlim), color = 'k', linestyle = 'dashed')
    ax[2].set_ylabel('$d$ [$\mu$m]')
    ax[2].grid()
    ax[2].text(x = max(f_interval) + 0.5, y = min(dlim) + 0.25 * (max(dlim) - min(dlim)), s = r'$%.2f +/- %.2f \mu$m' % (1e6*mean_fixed_thickness['d_weighted'], 1e6*mean_fixed_thickness['d_err_weighted']), color = 'r')
    ax[2].text(x = max(f_interval) + 0.5, y = min(dlim) + 0.05 * (max(dlim) - min(dlim)), s = r'$%.2f +/- %.2f \mu$m' % (1e6*statistic_average, 1e6*statistic_average_error), color = 'b')
    ax[2].set_ylim(min(dlim), max(dlim))
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_ref'] - mean['Ef_ref_err']), y2 = abs(mean['Ef_ref'] + mean['Ef_ref_err']), color = 'k', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_ref']), 'k-', label = 'ref')
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_sam0'] - mean['Ef_sam0_err']), y2 = abs(mean['Ef_sam0'] + mean['Ef_sam0_err']), color = 'r', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_sam0']), 'r-', label = 'main')
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_sam1'] - mean['Ef_sam1_err']), y2 = abs(mean['Ef_sam1'] + mean['Ef_sam1_err']), color = 'b', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_sam1']), 'b-', label = 'echo')
    if len(f_interval) == 2:
        ax[3].vlines(x = min(f_interval), ymin = 1e-5, ymax = 1.5, color = 'k', linestyle = 'dashed')
        ax[3].vlines(x = max(f_interval), ymin = 1e-5, ymax = 1.5, color = 'k', linestyle = 'dashed')
        ax[3].annotate('thickness interval', xy = (max(f_interval), 3e-2), xytext = (max(f_interval) + 0.5, 3e-5), arrowprops = dict(arrowstyle = '->', linewidth = lw_plot, color = 'k'), color = 'k')
    ax[3].set_yscale('log', nonpositive='clip')
    ax[3].set_xlabel('$f$ [THz]')
    ax[3].set_ylabel('$|E(f)|$ [norm.]')
    ax[3].legend(loc = 'lower left')
    ax[3].grid()
    ax[3].set_xlim(min(flim), max(flim))
    ax[3].set_ylim(1e-5, 1.5)
    if export == True:
        plt.savefig(r'%s\nkd.png' % path, dpi = 600, bbox_inches = 'tight')
    plt.show()
    return


def plot_nkd_weighted3(f, mean, mean_fixed_thickness, statistic_average, statistic_average_error, flim, nlim, klim, dlim, f_interval, title, export, path):
    lw_plot = 1

    size = 150 # mm
    aspectratio = 0.7#(1 + np.sqrt(5))/2
    plt.clf()
    fig = plt.figure(figsize = (size / 25.4, 1/aspectratio * size / 25.4))     
    ax = fig.subplots(nrows = 4, sharex = True)
    ax[0].fill_between(x = f, y1 = mean['n'] - mean['n_err'], y2 = mean['n'] + mean['n_err'], color = 'k', alpha = 0.5)
    ax[0].plot(f, mean['n'], 'k-', label = 'variable thickness')
    ax[0].fill_between(x = f, y1 = mean_fixed_thickness['n'] - mean_fixed_thickness['n_err'], y2 = mean_fixed_thickness['n'] + mean_fixed_thickness['n_err'], color = 'r', alpha = 0.5)
    ax[0].plot(f, mean_fixed_thickness['n'], 'r-', label = 'fixed thickness')
    ax[0].set_ylim(min(nlim), max(nlim))
    ax[0].set_ylabel('$n$')
    ax[0].grid()
    ax[0].legend(loc = 'upper left')
    ax[0].set_title(title)
    ax[1].fill_between(x = f, y1 = mean['k'] - mean['k_err'], y2 = mean['k'] + mean['k_err'], color = 'k', alpha = 0.5)
    ax[1].plot(f, mean['k'], 'k-')
    ax[1].fill_between(x = f, y1 = mean_fixed_thickness['k'] - mean_fixed_thickness['k_err'], y2 = mean_fixed_thickness['k'] + mean_fixed_thickness['k_err'], color = 'r', alpha = 0.5)
    ax[1].plot(f, mean_fixed_thickness['k'], 'r-')
    ax[1].set_ylabel('$k$')
    ax[1].set_ylim(min(klim), max(klim))
    ax[1].grid()
    ax[2].fill_between(x = f, y1 = 1e6*(mean['d'] - mean['d_err']), y2 = 1e6*(mean['d'] + mean['d_err']), color = 'k', alpha = 0.5)
    ax[2].plot(f, 1e6*mean['d'], 'k-')
    ax[2].fill_between(x = f, y1 = 1e6*np.ones(len(f))*(mean_fixed_thickness['d_weighted'] - mean_fixed_thickness['d_err_weighted']), y2 = 1e6*(mean_fixed_thickness['d_weighted'] + mean_fixed_thickness['d_err_weighted']), color = 'r', alpha = 0.5)
    ax[2].plot(f, 1e6*np.ones(len(f))*mean_fixed_thickness['d_weighted'], 'r-')
    if len(f_interval) == 2:
        ax[2].vlines(x = min(f_interval), ymin = min(dlim), ymax = max(dlim), color = 'k', linestyle = 'dashed')
        ax[2].vlines(x = max(f_interval), ymin = min(dlim), ymax = max(dlim), color = 'k', linestyle = 'dashed')
    ax[2].set_ylabel('$d$ [$\mu$m]')
    ax[2].grid()
    ax[2].text(x = min(f_interval) + 0.05, y = min(dlim) + 0.25 * (max(dlim) - min(dlim)), s = r'$%.2f +/- %.2f \mu$m' % (1e6*mean_fixed_thickness['d_weighted'], 1e6*mean_fixed_thickness['d_err_weighted']), color = 'r')
    #ax[2].text(x = max(f_interval) + 0.5, y = min(dlim) + 0.05 * (max(dlim) - min(dlim)), s = r'$%.2f +/- %.2f \mu$m' % (1e6*statistic_average, 1e6*statistic_average_error), color = 'b')
    ax[2].set_ylim(min(dlim), max(dlim))
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_ref'] - mean['Ef_ref_err']), y2 = abs(mean['Ef_ref'] + mean['Ef_ref_err']), color = 'k', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_ref']), 'k-', label = 'ref')
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_sam0'] - mean['Ef_sam0_err']), y2 = abs(mean['Ef_sam0'] + mean['Ef_sam0_err']), color = 'r', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_sam0']), 'r-', label = 'main')
    ax[3].fill_between(x = f, y1 = abs(mean['Ef_sam1'] - mean['Ef_sam1_err']), y2 = abs(mean['Ef_sam1'] + mean['Ef_sam1_err']), color = 'b', alpha = 0.5)
    ax[3].plot(f, abs(mean['Ef_sam1']), 'b-', label = 'echo')
    if len(f_interval) == 2:
        ax[3].vlines(x = min(f_interval), ymin = 1e-5, ymax = 1.5, color = 'k', linestyle = 'dashed')
        ax[3].vlines(x = max(f_interval), ymin = 1e-5, ymax = 1.5, color = 'k', linestyle = 'dashed')
        ax[3].annotate('thickness interval', xy = (max(f_interval), 3e-2), xytext = (max(f_interval) + 0.5, 3e-5), arrowprops = dict(arrowstyle = '->', linewidth = lw_plot, color = 'k'), color = 'k')
    ax[3].set_yscale('log', nonpositive='clip')
    ax[3].set_xlabel('$f$ [THz]')
    ax[3].set_ylabel('$|E(f)|$ [norm.]')
    ax[3].legend(loc = 'lower left')
    ax[3].grid()
    ax[3].set_xlim(min(flim), max(flim))
    ax[3].set_ylim(1e-5, 1.5)
    if export == True:
        plt.savefig(r'%s\nkd.png' % path, dpi = 600, bbox_inches = 'tight')
    plt.show()
    return
