# -*- coding: utf-8 -*-
"""
Created on Thu May 28 13:11:40 2020

@author: Andreas Gebauer
"""

"""
Pulses
"""

import numpy as np
import scipy.interpolate
import scipy.integrate
import scipy.special

import AG_file_operations_m1_website as AGf

###############################################################################
"""
Fourier and Jacobi transforms
"""

def t2w_transform(t, Et):
    """
    FFT from time domain to angular frequency domain
    If ifftshift is not used, Ew.real and Ew.imag exhibit rapid oscillations at the frequency peaks
    that are not present in the analytical fourier transform of e.g. a gaussian pulse. Such an 
    oscillation in w-space corresponds to a linear phase, i.e. a temporal shift (shift-theorem).
    ifftshift splits the pulse into a halfpulse at tmin and a halfpulse at tmax.
    """
    w = 2 * np.pi * np.fft.fftshift(np.fft.fftfreq(n = len(t), d = t[2] - t[1]))
    Ew = np.sqrt(len(w))*np.fft.fftshift(np.fft.fft(np.fft.ifftshift(Et), norm = 'ortho'))
    return w, Ew

def w2t_transform(w, Ew):
    """
    FFT from angular frequency domain to time domain
    """
    t = 2 * np.pi * np.fft.fftshift(np.fft.fftfreq(n = len(w), d = w[2] - w[1]))
    Et = 1/np.sqrt(len(t)) * np.fft.fftshift(np.fft.ifft(np.fft.ifftshift(Ew), norm = 'ortho'))
    return t, Et

def frequency_to_wavelength(f, I):
    c0 = 299792458
    wavelength = c0 / f
    I_l = I *  f **2 / c0
    return np.flip(wavelength), np.flip(I_l)

def wavelength_to_frequency(wavelength, I):
    c0 = 299792458
    f = c0 / wavelength
    I_f = I * wavelength**2 / c0
    return np.flip(f), np.flip(I_f)

def time_to_bandwidth(dt, pulseform):
      """
      dt in fs
      result dE in eV
      """
      e = 1.602176634e-19 
      h = 6.62607015e-34
      if pulseform == 'gaussian':
            dE =  2 * np.log(2)/np.pi * h/e * 1/(dt * 1e-15)
      else:
            raise ValueError('time_to_bandwidth: pulseform %s is not implemented.' % pulseform)
      return dE
            
def bandwidth_to_time(dE, pulseform):
      """
      dE in eV
      result dt in fs
      """
      e = 1.602176634e-19 
      h = 6.62607015e-34
      if pulseform == 'gaussian':
            dt = 2 * np.log(2)/np.pi * h/e * 1/dE * 1e15
      else:
            raise ValueError('bandwidth_to_time: pulseform %s is not implemented.' % pulseform)
      return dt
            

def nudft(x, y, omega, iflag=-1):
    """
    Non-Uniform Direct Fourier Transform
    (x, y) = sampled data, omega = new axis, iflag = -1 for FT, iflag = 1 for IFT 
    """
    sign = -1 if iflag < 0 else 1
    Ef = (1 / len(x)) * np.dot(y, np.exp(sign * 1j * omega * x[:, np.newaxis]))
    return Ef

###############################################################################
"""
cos2-pulses and -pulse trains in time domain
"""

def cos2_pulse(t, omega, T, CEP, E0):
      """
      Pulse with cos²-envelope. Note that if cos2_pulse is applied as a vector potential,
      electric field (cos2_vectorpotential_pulse) corresponds to its derivative and E0 then corresponds
      to vector potential amplitude.
      
      t: time (can be float or numpy-array)
      omega: central angular frequency (float)
      T: pulse duration (float); T corresponds to full vectorpotential-pulse duration and not to FWHM!
      FWHM (field) = 2 * T/pi * arccos(2^(-1/2)) = T/2
      FWHM (intensity) = 2 * T/pi * arccos(2^(-1/4)) = T * 0.36406
      T = 2 * FWHM_field = 2.7468 * FWHM_intensity
      CEP: Carrier-envelope-phase
      E0: electric field amplitude (float)
      """
      return E0 * np.cos(omega * t + CEP) * (np.cos(np.pi / T *t))**2 * (np.heaviside(t + T/2, 1)-np.heaviside(t - T/2, 1))

def cos2_pulse_envelope(t, T, E0):
      """
      Envelope of pulse with cos²-envelope. Note that if cos2_pulse is applied as a vector potential,
      electric field (cos2_dpulsedt) corresponds to its derivative and E0 then corresponds
      to vector potential amplitude.
      
      t: time (can be float or numpy-array)
      T: pulse duration (float); T corresponds to full vectorpotential-pulse duration and not to FWHM!
      E0: electric field amplitude (float)
      """
      return E0 * (np.cos(np.pi / T *t))**2 * (np.heaviside(t + T/2, 1)-np.heaviside(t - T/2, 1))

def cos2_dpulsedt(t, omega, T, CEP, E0):
      """
      Derivative of cos2_pulse. Can be used if cos2_pulse is used as a vector potential.
      t: time (can be float or numpy-array)
      omega: central angular frequency (float)
      T: pulse duration (float); T corresponds to full vectorpotential-pulse duration and not to FWHM!
      CEP: Carrier-envelope-phase
      E0: electric field amplitude (float)
      """
      return -E0*(np.cos(np.pi/T*t)**2*np.cos(omega * t + CEP)-2*np.pi/(T * omega) * np.cos(np.pi/T * t) * np.sin(np.pi/T * t) * np.sin(omega * t + CEP)) * (np.heaviside(t + T/2, 1) - np.heaviside(t - T/2, 1))

def cos2_pulse_vectorpotential(t, omega, T, CEP, E0):
      """
      A(t) = - Integral(E(t) dt)
      E(t) following definition of cos2_pulse
      """
      return E0/4 * (-np.sin((omega - 2*np.pi/T)*t + CEP)/(omega - 2*np.pi/T) -np.sin((omega + 2*np.pi/T)*t + CEP)/(omega + 2*np.pi/T) -2*np.sin(omega*t + CEP)/omega -np.sin(omega*T/2+CEP) * (2/omega - 1/(omega - 2*np.pi/T) - 1/(omega + 2*np.pi/T))) * (np.heaviside(t + T/2, 1)-np.heaviside(t - T/2, 1)) +E0/2 * np.sin(omega*T/2+CEP) * (-2/omega + 1/(omega - 2*np.pi/T) + 1/(omega + 2*np.pi/T)) * np.heaviside(t - T/2, 1)

def cos2_pulsetrain(t, T, omega1, omega2, E0, E1, E2, CEP1, CEP2):
      """
      Pulse train with cos²-envelope. Note that if cos2_pulsetrain is applied as a vector potential,
      electric field (cos2_dpulsetraindt) corresponds to its derivative and E0 then corresponds
      to vector potential amplitude.
      
      t: time (can be float or numpy-array)
      omega1: angular frequency of first spectral line (float)
      omega2: angular frequency of second spectral line (float)
      T: pulse duration (float); T corresponds to full vectorpotential-pulse duration and not to FWHM!
      E0: electric field amplitude (float)
      E1, E2: relative electric field strength of the two harmonics. Values make sense between -1 and 1. If both are =1 this function corresponds to its envelope function cos2_pulsetrain_envelope
      """
      return E0 * (E1 * np.cos(omega1 * t + CEP1) + E2 * np.cos(omega2 * t + CEP2)) * (np.cos(np.pi / T *t))**2 * (np.heaviside(t + T/2, 1)-np.heaviside(t - T/2, 1))


def cos2_pulsetrain_envelope(t, T, omega1, omega2, E0):
      """
      Envelope of pulse train with cos²-envelope. Can be used if cos2_pulsetrain is used as a vector 
      potential.
      
      t: time (can be float or numpy-array)
      omega1: angular frequency of first spectral line (float)
      omega2: angular frequency of second spectral line (float)
      T: pulse duration (float); T corresponds to full vectorpotential-pulse duration and not to FWHM!
      E0: electric field amplitude (float)
      """
      omega = (omega1 + omega2) / 2
      return E0 * np.cos(np.pi/T * t)**2 * (np.cos((omega - omega1) * t) + np.cos((omega - omega2) * t)) * (np.heaviside(t + T/2, 1) - np.heaviside(t - T/2, 1))       

def cos2_dpulsetraindt(t, T, omega1, omega2, E0):
      """
      Derivative of pulse train with cos²-envelope. Note that if cos2_pulsetrain is applied as a vector potential,
      electric field (cos2_dpulsetraindt) corresponds to its derivative and E0 then corresponds
      to vector potential amplitude.
      
      t: time (can be float or numpy-array)
      omega1: angular frequency of first spectral line (float)
      omega2: angular frequency of second spectral line (float)
      T: pulse duration (float); T corresponds to full vectorpotential-pulse duration and not to FWHM!
      E0: electric field amplitude (float)
      """
      omega = (omega1 + omega2) / 2
      return -E0 * (np.cos(np.pi/T * t)**2 * (np.cos(omega * t) * (np.cos((omega - omega1) * t) + np.cos((omega - omega2) * t)) - np.sin(omega * t) * ((omega-omega1)/omega * np.sin((omega - omega1) * t) + (omega - omega2)/omega * np.sin((omega - omega2) * t))) - np.cos(np.pi/T * t) * np.sin(np.pi/T * t) * 2*np.pi/(omega * T) * np.sin(omega * t) * (np.sin((omega - omega1) * t) + np.sin((omega - omega2) * t))) * (np.heaviside(t + T/2, 1) - np.heaviside(t - T/2, 1))

def cos2_pulsetrain_vectorpotential(t, T, omega1, omega2, E0, E1, E2):
    """
    A(t) = - Integral(E(t) dt)
    E(t) following definition of cos2_pulsetrain
    """
    return E0 * (cos2_pulse_vectorpotential(t = t, omega = omega1, T = T, CEP = 0.0, E0 = E1) + cos2_pulse_vectorpotential(t = t, omega = omega2, T = T, CEP = 0.0, E0 = E2))

###############################################################################
"""
cos2-pulses and -pulse trains in frequency domain
"""

def cos2_pulse_spectral(omega_axis, omega, T, CEP, E0):
    """
    Notizbuch S. 47
    Numpy definition of sinc: sinc(x) = sin(pi * x) / (pi * x) !!!
    """
    E_w = E0 * T / (8 * np.sqrt(2*np.pi)) * (2*np.sinc(1/np.pi*(omega_axis - omega)*T/2)*np.exp(1j*CEP) + 2*np.sinc(1/np.pi*(omega_axis + omega)*T/2)*np.exp(-1j*CEP) + np.sinc(1/np.pi*(omega_axis - omega - 2*np.pi/T)*T/2)*np.exp(1j*CEP) + np.sinc(1/np.pi*(omega_axis + omega + 2*np.pi/T)*T/2)*np.exp(-1j*CEP) + np.sinc(1/np.pi*(omega_axis - omega + 2*np.pi/T)*T/2)*np.exp(1j*CEP) + np.sinc(1/np.pi*(omega_axis + omega - 2*np.pi/T)*T/2)*np.exp(-1j*CEP))
    return E_w

def cos2_pulsetrain_spectral(omega_axis, omega1, omega2, T, E0, E1, E2, CEP1, CEP2):
    E_w = E0 * (cos2_pulse_spectral(omega_axis = omega_axis, omega = omega1, T = T, CEP = CEP1, E0 = E1) + cos2_pulse_spectral(omega_axis = omega_axis, omega = omega2, T = T, CEP = CEP2, E0 = E2))
    return E_w


###############################################################################
"""
gaussian pulses and pulse train in time domain
"""
def gaussian_chirped_pulse(t, omega, T, chirp, CEP, E0):
      """
      returns a numpy-array containing the vectorpotential envelope of the EUV pulse following the definition of
      Wollenhaupt, p. 1055 (Springer Handbook of Lasers and Optics)
      -> T corresponds to FWHM of Fourier-limited pulse!
      """
      beta = T**2/(8*np.log(2))
      gamma = 1 + chirp**2/(4*beta**2)
      alpha = chirp/(8*beta**2*gamma)
      epsilon = 1/2 * np.arctan2(chirp, 2*beta) # additional CEP that is introduced by chirp
      return E0 * gamma**(-1/4) * np.exp(-t**2/(4*beta*gamma)) * np.cos(omega * t + alpha*t**2 - epsilon + CEP)

def gaussian_chirped_pulse_envelope(t, omega, T, chirp, E0):
    """
    returns a numpy-array containing the vectorpotential envelope of the EUV pulse following the definition of
    Wollenhaupt, p. 1055 (Springer Handbook of Lasers and Optics)
    -> T corresponds to FWHM!
    """
    beta = T**2/(8*np.log(2))
    gamma = 1 + chirp**2/(4*beta**2)
    alpha = chirp/(8*beta**2*gamma)
    pulse_envelope = E0 * gamma**(-1/4) * np.exp(-t**2/(4*beta*gamma))
    #excitation_energy_time = np.array([(omega_EUV+2*alpha*(t+delay_IREUV)) for t in t_axis])
    return pulse_envelope

def gaussian_pulsetrain(t, T, omega1, omega2, chirp, CEP, E0):
      """
      -> T corresponds to FWHM!
      chirp definition equivalent to gaussian_chirped_pulse
      This definition only excepts E1 = E2 = E0/2 because otherwise an envelope cannot be defined.
      If E1 != E2 is desired, use superposition of two chirped gaussian pulses.
      """
      omega = (omega1 + omega2) / 2
      domega = (omega2 - omega1) / 2
      beta = T**2/(8*np.log(2))
      gamma = 1 + chirp**2/(4*beta**2)
      alpha = chirp/(8*beta**2*gamma)
      epsilon = 1/2 * np.arctan2(chirp, 2*beta)
      return E0 * gamma**(-1/4) * np.exp(-t**2 / (4*beta*gamma)) * np.cos(domega * t) * np.cos(omega * t + alpha * t**2 - epsilon + CEP)

def gaussian_pulsetrain_envelope(t, T, omega1, omega2, chirp, CEP, E0):
      """
      -> T corresponds to FWHM!
      chirp definition equivalent to gaussian_chirped_pulse
      This definition only excepts E1 = E2 = E0/2 because otherwise an envelope cannot be defined.
      If E1 != E2 is desired, use superposition of two chirped gaussian pulses.
      Carrier wave: np.cos(omega * t + alpha * t**2 - epsilon + CEP)
      """
      #omega = (omega1 + omega2) / 2
      domega = (omega2 - omega1) / 2
      beta = T**2/(8*np.log(2))
      gamma = 1 + chirp**2/(4*beta**2)
      #alpha = chirp/(8*beta**2*gamma)
      #epsilon = 1/2 * np.arctan2(chirp, 2*beta)
      return E0 * gamma**(-1/4) * np.exp(-t**2 / (4*beta*gamma)) * np.cos(domega * t)

def gaussian_pulse_vectorpotential(t, omega, T, E0):
    """
    chirp is not implemented yet
    """
    beta = T**2/(8*np.log(2))
    return E0 * np.sqrt(beta * np.pi)/2 * np.exp(-beta*omega**2) * (2 - scipy.special.erf((t - 2*1j*beta*omega) / (2*np.sqrt(beta))).real - scipy.special.erf((t + 2*1j*beta*omega) / (2*np.sqrt(beta))).real)




###############################################################################
"""
gaussian pulses and pulse train in frequency domain
"""
def gaussian_chirped_pulse_frequencydomain(w, delay, omega, T, chirp, CEP, E0):
      """
      w is angular frequency axis
      omega is central frequency
      returns a numpy-array containing the vectorpotential envelope of the EUV pulse following the definition of
      Wollenhaupt, p. 1055 (Springer Handbook of Lasers and Optics)
      -> T corresponds to FWHM of Fourier-limited pulse!
      """
      #return E0 * T/2 * np.sqrt(np.pi/(2*np.log(2))) * (np.exp(- T**2 / (8 * np.log(2)) * (w - omega)**2 + 1j * delay * (w - omega) - 1j*(chirp/2 * (w - omega)**2)) + np.exp(- T**2 / (8 * np.log(2)) * (w + omega)**2 - 1j * delay * (w + omega) + 1j*(chirp/2 * (w + omega)**2)) )
      return E0 * T/2 * np.sqrt(np.pi/(2*np.log(2))) * (np.exp(- T**2 / (8 * np.log(2)) * (w - omega)**2 - 1j * delay * w - 1j*(chirp/2 * (w - omega)**2)) + np.exp(- T**2 / (8 * np.log(2)) * (w + omega)**2 - 1j * delay * w + 1j*(chirp/2 * (w + omega)**2)) )

###############################################################################
"""
Auxilary functions
"""

def super_gauss(x, x0, FWHM, n):
    sigma = FWHM / 2 * (2*np.log(2))**(-1/n)
    return np.exp(-abs(x - x0)**n / (2 * sigma**n))

def Kaiser_Bessel_window(t, T, beta):
    """
    centered around t = 0, full width of T
    """
    return np.nan_to_num(scipy.special.i0(beta * np.sqrt(1 - (2*t/T)**2)) / scipy.special.i0(beta))

def Kaiser_Bessel_window_spectral(omega, T, beta):
    return np.where(beta >= abs(omega*T/2), 2*T/scipy.special.i0(beta) * np.nan_to_num(np.sinh(np.sqrt(beta**2 - (omega*T/2)**2))/np.sqrt(beta**2 - (omega*T/2)**2)), 2*T/scipy.special.i0(beta) * np.nan_to_num(np.sin(np.sqrt((omega*T/2)**2 - beta**2))/np.sqrt((omega*T/2)**2 - beta**2)))

def rectangular_window(t, t0, T):
    return np.heaviside((t - t0 + T/2), 0) + np.heaviside(-(t - t0 - T/2), 0) - 1

def symmetric_zeropadding(x_in, y_in, Nx_out):
    if Nx_out > len(x_in):
        dx = np.mean(np.diff(x_in))
        zeros_half = np.zeros(int((Nx_out - len(x_in))/2))
        x_zeros_left = np.linspace(min(x_in) - dx * len(zeros_half),  min(x_in) - dx, len(zeros_half))
        x_zeros_right = np.linspace(max(x_in) + dx, max(x_in) + dx * len(zeros_half), len(zeros_half))
        x_out = np.concatenate([x_zeros_left, x_in, x_zeros_right])
        y_out = np.concatenate([zeros_half, y_in, zeros_half])
    else:
        x_out = x_in
        y_out = y_in
    return x_out, y_out

###############################################################################
"""
Special cases
"""

def EUV_pulse_streaking(t, IREUV_delay, type_of_EUV_pulse, EUV_parameters):
    """
    Pulse duration corresponds to full pulse, NOT FWHM! (Warum? Noch aktuell?)
    Electric field of EUV-pulse is calculated with the definitions used for vector potential in TDSE_FFT_Numerov_v2.6.py (noch aktuell?)
    """    
    if type_of_EUV_pulse == 'SAP':
        if len(EUV_parameters) != 4:
              raise ValueError('EUV_pulse_streaking: EUV_parameters do not match the type_of_EUV_pulse = SAP')
        [omega_EUV_SAP, pulseduration_EUV_SAP, chirp_EUV_SAP, fieldstrength_EUV_SAP] = EUV_parameters
        E_EUV = fieldstrength_EUV_SAP * np.cos((t+IREUV_delay) * np.pi / pulseduration_EUV_SAP)**2 * np.cos(omega_EUV_SAP * (t+IREUV_delay)) * (np.heaviside((t+IREUV_delay) + pulseduration_EUV_SAP/2, 1) - np.heaviside((t+IREUV_delay) - pulseduration_EUV_SAP/2, 1))
    elif type_of_EUV_pulse == 'APT':
        if len(EUV_parameters) != 7:
              raise ValueError('EUV_pulse_streaking: EUV_parameters do not match the type_of_EUV_pulse = APT')
        [omega_EUV_1_APT, I_EUV_1_APT, omega_EUV_2_APT, I_EUV_2_APT, phi_APT, pulseduration_EUV_APT, fieldstrength_EUV_APT] = EUV_parameters
        E_EUV = fieldstrength_EUV_APT * np.cos((t+IREUV_delay) * np.pi / pulseduration_EUV_APT)**2 * (I_EUV_1_APT * np.cos(omega_EUV_1_APT * (t + IREUV_delay)) + I_EUV_2_APT * np.cos(omega_EUV_2_APT * (t + IREUV_delay) + phi_APT)) * (np.heaviside((t+IREUV_delay) + pulseduration_EUV_APT/2, 1) - np.heaviside((t+IREUV_delay) - pulseduration_EUV_APT/2, 1))
    elif type_of_EUV_pulse == 'File':
        if len(EUV_parameters) != 3:
              raise ValueError('EUV_pulse_streaking: EUV_parameters do not match the type_of_EUV_pulse = File')
        [EUV_timedomain_file, pulseduration_EUV_file, fieldstrength_EUV_file] = EUV_parameters
        t_axis_EUV, E0_EUV, E_EUV, t_EUV_min, t_EUV_max, w0 = AGf.load_EUV_file(EUV_timedomain_file)
        E0_EUV = E0_EUV/max(abs(E0_EUV))
        E_EUV = E_EUV/max(abs(E_EUV))
        if max(t) > max(t_axis_EUV) or min(t) < min(t_axis_EUV):
            raise ValueError('EUV_pulse_vectorpotential_envelope: t_axis exceeds the range of the EUV pulse file')
        E_EUV_int = scipy.interpolate.interp1d(t_axis_EUV, E_EUV, kind='cubic')
        E_EUV = fieldstrength_EUV_file * E_EUV_int(t+IREUV_delay)
    else:
        raise ValueError('EUV_pulse_vectorpotential_envelope: type_of_EUV_pulse %s is not yet implemented.' % type_of_EUV_pulse)
    return E_EUV

def EUV_pulse_vectorpotential_envelope(t_axis, delay_IREUV, EUV_parameters):
    """
    Pulse duration corresponds to full pulse, NOT FWHM! (noch aktuell?)
    For SAP and APT envelope of vector potential is calculated exactly without slowly-varying envelope approximation.
    Instead EUV_pulse (see above) is calculated by the derivative of A_EUV(t) if necessary.
    Center wavelength oscillation is a sin-function.
    """    
    if EUV_parameters['type_of_EUV_pulse'] == 'SAP':
        if EUV_parameters['pulse_shape'] == 'cos2':
              A_EUV = cos2_pulse_envelope(t = t_axis + delay_IREUV, T = EUV_parameters['pulseduration_EUV'], E0 = EUV_parameters['fieldstrength_EUV'])
        elif EUV_parameters['pulse_shape'] == 'gaussian':
              A_EUV = gaussian_chirped_pulse_envelope(t = t_axis + delay_IREUV, omega = EUV_parameters['omega_EUV'], T = EUV_parameters['pulseduration_EUV'], chirp = EUV_parameters['chirp_EUV'], E0 = EUV_parameters['fieldstrength_EUV'])
        else:
              raise ValueError('EUV_pulse_vectorpotential_envelope: pulse_shape = %s in EUV_parameters is not yet implemented' % EUV_parameters['pulse_shape'])
    elif EUV_parameters['type_of_EUV_pulse'] == 'APT':
        if EUV_parameters['pulse_shape'] == 'cos2':
              A_EUV = cos2_pulsetrain_envelope(t = t_axis + delay_IREUV, T = EUV_parameters['pulseduration_EUV'], omega1 = EUV_parameters['omega_EUV_1'], omega2 = EUV_parameters['omega_EUV_2'], E0 = EUV_parameters['fieldstrength_EUV'])
        elif EUV_parameters['pulse_shape'] == 'gaussian':
              A_EUV = gaussian_pulsetrain_envelope(t = t_axis + delay_IREUV, T = EUV_parameters['pulseduration_EUV'], omega1 = EUV_parameters['omega_EUV_1'], omega2 = EUV_parameters['omega_EUV_2'], chirp = EUV_parameters['chirp_EUV'], CEP = 0.0, E0 = EUV_parameters['fieldstrength_EUV'])
        else:
              raise ValueError('EUV_pulse_vectorpotential_envelope: pulse_shape = %s in EUV_parameters is not yet implemented' % EUV_parameters['pulse_shape'])
    elif EUV_parameters['type_of_EUV_pulse'] == 'File':
        t_axis_EUV, E0_EUV, E_EUV, t_EUV_min, t_EUV_max, w0 = AGf.load_EUV_file(EUV_parameters['EUV_timedomain_file'])
        E0_EUV = E0_EUV/max(abs(E0_EUV))
        E_EUV = E_EUV/max(abs(E_EUV))
        if max(t_axis) > max(t_axis_EUV) or min(t_axis) < min(t_axis_EUV):
            raise ValueError('EUV_pulse_vectorpotential_envelope: t_axis exceeds the range of the EUV pulse file')
        E0_EUV_int = scipy.interpolate.interp1d(t_axis_EUV, E0_EUV, kind='cubic')
        A_EUV = EUV_parameters['fieldstrength_EUV'] * E0_EUV_int(t_axis+delay_IREUV)
    else:
        raise ValueError('EUV_pulse_vectorpotential_envelope: type_of_EUV_pulse %s in EUV_parameters is not yet implemented.' % EUV_parameters['type_of_EUV_pulse'])
    return A_EUV

###############################################################################
"""
Second-order processes

This section is not finished and should not yet be used!
"""

def SFG_chirped_gaussians_timedomain(t, tau_1, omega_1, T_1, chirp_1, CEP_1, E_1, tau_2, omega_2, T_2, chirp_2, CEP_2, E_2):
    beta_1 = T_1**2/(8*np.log(2))
    gamma_1 = 1 + chirp_1**2/(4*beta_1**2)
    alpha_1 = chirp_1/(8*beta_1**2*gamma_1)
    epsilon_1 = 1/2 * np.arctan2(chirp_1, 2*beta_1)
    beta_2 = T_2**2/(8*np.log(2))
    gamma_2 = 1 + chirp_2**2/(4*beta_2**2)
    alpha_2 = chirp_2/(8*beta_2**2*gamma_2)
    epsilon_2 = 1/2 * np.arctan2(chirp_2, 2*beta_2)
    print(beta_1, gamma_1, alpha_1, epsilon_1)
    print(beta_2, gamma_2, alpha_2, epsilon_2)
    if omega_1 == omega_2 and T_1 == T_2 and chirp_1 == chirp_2:
        """
        Second harmonic generation (special case)
        """
        E_SFG_t = E_1 * E_2 / (2 * np.sqrt(gamma_1)) * np.exp(-(t - tau_1)**2 / (2*beta_1*gamma_1)) * (np.cos(2*omega_1 * (t - (tau_1 + tau_2)/2) + 2*alpha_1 * (t - (tau_1 + tau_2)/2)**2 + alpha_1 * (tau_1 - tau_2)**2 - 2*epsilon_1 + CEP_1 + CEP_2) + np.cos(-2*alpha_1 * (tau_1 - tau_2) * t - omega_1/2 * (tau_1 - tau_2) + alpha_1 * (tau_1**2 - tau_2**2) + CEP_1 - CEP_2))
    else:
        C1 = E_1 * E_2 / ( gamma_1**(1/4) * gamma_2**(1/4)) * np.exp(-(tau_1 - tau_2)**2 / (4*(beta_1*gamma_1 + beta_2*gamma_2)))
        C2 = (beta_1 * gamma_1 + beta_2 * gamma_2) / (4*beta_1*beta_2*gamma_1*gamma_2)
        tau0 = (beta_2*gamma_2*tau_1 + beta_1*gamma_1*tau_2) / (beta_1*gamma_1 + beta_2*gamma_2)
        wp = omega_1 + omega_2
        ap = alpha_1 + alpha_2
        tp = (alpha_1*tau_1 + alpha_2*tau_2) / (alpha_1 + alpha_2)
        ep = alpha_1*alpha_2 / (alpha_1 + alpha_2) * (tau_1 - tau_2)**2 - omega_1*tau_1 - omega_2*tau_2 - epsilon_1 - epsilon_2 + CEP_1 + CEP_2
        wm = omega_1 - omega_2
        am = alpha_1 - alpha_2
        tm = (alpha_1*tau_1 - alpha_2*tau_2) / (alpha_1 - alpha_2)
        em = -alpha_1*alpha_2 / (alpha_1 - alpha_2) * (tau_1 + tau_2)**2 - omega_1*tau_1 + omega_2*tau_2 - epsilon_1 + epsilon_2 + CEP_1 - CEP_2
        E_SFG_t = C1 * np.exp(-C2 * (t - tau0)**2) #* (np.cos(wp*t + ap*(t - tp)**2 + ep) + np.cos(wm*t + am*(t - tm)**2 + em))
    return E_SFG_t


###############################################################################
"""
SPIDER
"""

def E_SPIDER_timedomain(t, omega0, T0, chirp0, CEP0, tau_doublepulse, tau_stretch, chirp_stretch):
    E1 = gaussian_chirped_pulse(t = t - tau_doublepulse/2, omega = omega0, T = T0, chirp = chirp0, CEP = CEP0, E0 = 1)
    E2 = gaussian_chirped_pulse(t = t + tau_doublepulse/2, omega = omega0, T = T0, chirp = chirp0, CEP = CEP0, E0 = 1)
    E_stretch = gaussian_chirped_pulse(t = t - tau_stretch, omega = omega0, T = T0, chirp = chirp0 + chirp_stretch, CEP = CEP0, E0 = 1)
    E_SPIDER = E_stretch * (E1 + E2)
    return E_SPIDER

def generate_SPIDER_spectrum_chirped_gaussian(t, lambda0, T0, chirp0, CEP0, tau_doublepulse, tau_stretch, chirp_stretch):
    """
    t: time axis
    lambda0: central wavelength of input pulse
    T0: pulse duration of input pulse
    chirp0: GDD of input pulse
    CEP0: Carrier Envelope Phase of input pulse
    tau_doublepulse: delay between pulse copies introduced by etalon
    tau_stretch: delay between stretched pulse and center of mass of double pulse (delay stage)
    chirp_stretch: additional chirp due to glass block
    """
    c = 299792458
    omega0 = 2*np.pi * c/lambda0
    E_SPIDER_t = E_SPIDER_timedomain(t = t, omega0 = omega0, T0 = T0, chirp0 = chirp0, CEP0 = CEP0, tau_doublepulse = tau_doublepulse, tau_stretch = tau_stretch, chirp_stretch = chirp_stretch)    
    w, E_SPIDER_w = t2w_transform(t = t, Et = E_SPIDER_t)
    I_SPIDER_w = abs(E_SPIDER_w)**2 / max(abs(E_SPIDER_w)**2)
    E_fundamental_w = gaussian_chirped_pulse_frequencydomain(w = w, omega = omega0, T = T0, chirp = chirp0, CEP = CEP0, E0 = 1)
    I_fundamental_w = abs(E_fundamental_w)**2 / max(abs(E_fundamental_w)**2)
    f = w / (2*np.pi)
    
    cut_min_SPIDER = next(i for i in range(len(f)) if f[i] >= 500e12)
    cut_max_SPIDER = next(i for i in range(len(f)) if f[i] >= 1000e12)
    f_SPIDER = f[cut_min_SPIDER:cut_max_SPIDER]
    I_SPIDER_f = I_SPIDER_w[cut_min_SPIDER:cut_max_SPIDER]
    
    L_SPIDER, I_SPIDER_L = frequency_to_wavelength(f = f_SPIDER, I = I_SPIDER_f)
    L_SPIDER_new = np.linspace(min(L_SPIDER), max(L_SPIDER), 15000)
    I_SPIDER_L_int = scipy.interpolate.interp1d(L_SPIDER, I_SPIDER_L, kind = 'cubic')
    I_SPIDER_L_new = I_SPIDER_L_int(L_SPIDER_new)
    
    cut_min_fundamental = next(i for i in range(len(f)) if f[i] >= 200e12)
    cut_max_fundamental = next(i for i in range(len(f)) if f[i] >= 500e12)
    f_fundamental = f[cut_min_fundamental:cut_max_fundamental]
    I_fundamental_f = I_fundamental_w[cut_min_fundamental:cut_max_fundamental]
        
    L_fundamental, I_fundamental_L = frequency_to_wavelength(f = f_fundamental, I = I_fundamental_f)
    L_fundamental_new = np.linspace(min(L_fundamental), max(L_fundamental), 15000)
    I_fundamental_L_int = scipy.interpolate.interp1d(L_fundamental, I_fundamental_L, kind = 'cubic')
    I_fundamental_L_new = I_fundamental_L_int(L_fundamental_new)
    
    return L_SPIDER_new, I_SPIDER_L_new / max(abs(I_SPIDER_L_new)), L_fundamental_new, I_fundamental_L_new / max(abs(I_fundamental_L_new))

def generate_SPIDER_spectrum_general_pulse(omega, omega0, spectrum_f, phase_f, tau_doublepulse, tau_stretch, chirp_stretch):
    """
    omega:              angular frequency axis (not afrequency!). Omega axis needs to be symmetric to zero 
                        ranging from -omega_max to omega_max.
    omega0:             center frequency
    spectrum_f:         spectrum in frequency domain
    phase_f:            phase in frequency domain
    tau_doublepulse:    delay between pulse copies introduced by etalon
    tau_stretch:        delay between stretched pulse and center of mass of double pulse (delay stage)
    chirp_stretch:      additional chirp due to glass block
    """
    E_original_pulsecopy1_f = np.sqrt(spectrum_f) * np.exp(-1j * (phase_f - tau_doublepulse/2 * (omega - omega0)))
    E_original_pulsecopy2_f = np.sqrt(spectrum_f) * np.exp(-1j * (phase_f + tau_doublepulse/2 * (omega - omega0)))
    E_stretch_f = np.sqrt(spectrum_f) * np.exp(-1j * (phase_f - tau_stretch * (omega - omega0) + chirp_stretch/2 * (omega - omega0)**2))
    t, E_original_pulsecopy1_t = w2t_transform(w = omega, Ew = E_original_pulsecopy1_f)
    t, E_original_pulsecopy2_t = w2t_transform(w = omega, Ew = E_original_pulsecopy2_f)
    t, E_stretch_t = w2t_transform(w = omega, Ew = E_stretch_f)
    E_SPIDER_t = (E_original_pulsecopy1_t + E_original_pulsecopy2_t) * E_stretch_t
    w, E_SPIDER_w = t2w_transform(t = t, Et = E_SPIDER_t)
    f = w / (2*np.pi)
    I_SPIDER_f = abs(E_SPIDER_w)**2 / max(abs(E_SPIDER_w)**2)
    
    cut_min_SPIDER = next(i for i in range(len(f)) if f[i] >= 500e12)
    cut_max_SPIDER = next(i for i in range(len(f)) if f[i] >= 1000e12)
    f_SPIDER = f[cut_min_SPIDER:cut_max_SPIDER]
    I_SPIDER_f = I_SPIDER_f[cut_min_SPIDER:cut_max_SPIDER]
    
    L_SPIDER, I_SPIDER_L = frequency_to_wavelength(f = f_SPIDER, I = I_SPIDER_f)
    L_SPIDER_new = np.linspace(min(L_SPIDER), max(L_SPIDER), 15000)
    I_SPIDER_L_int = scipy.interpolate.interp1d(L_SPIDER, I_SPIDER_L, kind = 'cubic')
    I_SPIDER_L_new = I_SPIDER_L_int(L_SPIDER_new)
    
    cut_min_fundamental = next(i for i in range(len(omega)) if omega[i] / (2*np.pi) >= 200e12)
    cut_max_fundamental = next(i for i in range(len(omega)) if omega[i] / (2*np.pi) >= 500e12)
    f_fundamental = omega[cut_min_fundamental:cut_max_fundamental] / (2*np.pi)
    I_fundamental_f = spectrum_f[cut_min_fundamental:cut_max_fundamental]
        
    L_fundamental, I_fundamental_L = frequency_to_wavelength(f = f_fundamental, I = I_fundamental_f)
    L_fundamental_new = np.linspace(min(L_fundamental), max(L_fundamental), 15000)
    I_fundamental_L_int = scipy.interpolate.interp1d(L_fundamental, I_fundamental_L, kind = 'cubic')
    I_fundamental_L_new = I_fundamental_L_int(L_fundamental_new)
    return L_SPIDER_new, I_SPIDER_L_new / max(abs(I_SPIDER_L_new)), L_fundamental_new, I_fundamental_L_new / max(abs(I_fundamental_L_new))
    

def SPIDER_unit_conversion(mode, lambda_SPIDER, I_SPIDER, lambda_fundamental, I_fundamental, lambda0, shear, shift, tau_SG, Stdev_SG, smooth):
      if mode == 'to_SI':
            lambda_SPIDER = lambda_SPIDER * 1e-9
            I_SPIDER = I_SPIDER #* 1e9
            lambda_fundamental = lambda_fundamental * 1e-9
            I_fundamental = I_fundamental #* 1e9
            lambda0 = lambda0 * 1e-9
            shear = shear * 1e12
            shift = shift * 1e12
            tau_SG = tau_SG * 1e-15 
            Stdev_SG = Stdev_SG * 1e-15 
            smooth = smooth * 1e12
      elif mode == 'to_THz-fs':
            lambda_SPIDER = lambda_SPIDER * 1e9
            I_SPIDER = I_SPIDER * 1e-9
            lambda_fundamental = lambda_fundamental * 1e9
            I_fundamental = I_fundamental * 1e-9
            lambda0 = lambda0 * 1e9
            shear = shear * 1e-12
            shift = shift * 1e-12
            tau_SG = tau_SG * 1e15 
            Stdev_SG = Stdev_SG * 1e15 
            smooth = 1e-12
      else:
            raise ValueError('unit_conversion: mode %s is invalid.' % mode)
      return lambda_SPIDER, I_SPIDER, lambda_fundamental, I_fundamental, lambda0, shear, shift, tau_SG, Stdev_SG, smooth

def filter_SPIDER_spectrum(lambda_SPIDER, I_SPIDER, tau, Stdev_SG):
      f_SPIDER, I_f_SPIDER = wavelength_to_frequency(wavelength = lambda_SPIDER, I = I_SPIDER)
      f_SPIDER_new = np.linspace(-2*f_SPIDER[-1], 2*f_SPIDER[-1], 2**16)
      I_f_SPIDER_int = scipy.interpolate.interp1d(f_SPIDER, I_f_SPIDER, kind = 'cubic', bounds_error = False, fill_value = 0.0)
      I_f_SPIDER_new = I_f_SPIDER_int(f_SPIDER_new)
      E_f_SPIDER_new = np.sqrt(abs(I_f_SPIDER_new))
      
      t_SPIDER, E_SPIDER_t = w2t_transform(w = 2*np.pi*f_SPIDER_new, Ew = E_f_SPIDER_new)     
      SG = super_gauss(x = t_SPIDER, x0 = tau, FWHM = 2*Stdev_SG, n = 8)
      E_SPIDER_filtered_t = E_SPIDER_t * np.sqrt(SG)
      w_SPIDER, E_SPIDER_filtered_f = t2w_transform(t = t_SPIDER, Et = E_SPIDER_filtered_t)     
      return f_SPIDER_new, E_SPIDER_filtered_f

def retrieve_group_delay(f_SPIDER, I_SPIDER, tau, shear):
      phase_raw = np.unwrap(np.angle(I_SPIDER, deg = False))
      GD = (phase_raw + 2*np.pi*f_SPIDER * tau) / shear
      return GD

def calculate_pulse(lambda_fundamental, I_fundamental, f_SPIDER_shifted, phase):
      f_fundamental, I_f_fundamental = wavelength_to_frequency(wavelength = lambda_fundamental, I = I_fundamental)
      I_f_fundamental_int = scipy.interpolate.interp1d(f_fundamental, I_f_fundamental, kind = 'cubic', bounds_error = False, fill_value = 0.0)
      I_f_fundamental_new = I_f_fundamental_int(f_SPIDER_shifted)
      E_f_fundamental_new = np.sqrt(abs(I_f_fundamental_new))
      
      E_f = E_f_fundamental_new * np.exp(-complex(0,1) * phase)
      t, E_t = w2t_transform(w = 2*np.pi * f_SPIDER_shifted, Ew = E_f)
      return t, abs(E_t)**2    

def SPIDER_pulse_reconstruction(lambda_SPIDER, I_SPIDER, lambda_fundamental, I_fundamental, lambda0, shear, shift, tau_SG, Stdev_SG, smooth, unit_mode):
      if unit_mode == 'THz-fs':
            lambda_SPIDER, I_SPIDER, lambda_fundamental, I_fundamental, lambda0, shear, shift, tau_SG, Stdev_SG, smooth = SPIDER_unit_conversion(mode = 'to_SI', lambda_SPIDER = lambda_SPIDER, I_SPIDER = I_SPIDER, lambda_fundamental = lambda_fundamental, I_fundamental = I_fundamental, lambda0 = lambda0, shear = shear, shift = shift, tau_SG = tau_SG, Stdev_SG = Stdev_SG, smooth = smooth)
      f_SPIDER_AC, E_SPIDER_AC = filter_SPIDER_spectrum(lambda_SPIDER = lambda_SPIDER, I_SPIDER = I_SPIDER, tau = tau_SG, Stdev_SG = Stdev_SG)
      f_SPIDER_shifted = f_SPIDER_AC - shift
      i_center = next(i for i in range(len(f_SPIDER_shifted)) if f_SPIDER_shifted[i] >= 299792458/lambda0)
      GD = retrieve_group_delay(f_SPIDER = f_SPIDER_shifted, I_SPIDER = E_SPIDER_AC, tau = tau_SG, shear = shear)
      window = int(smooth / (f_SPIDER_AC[2] - f_SPIDER_AC[1]))
      if window != 0:
            GD = np.convolve(GD, np.ones(window)/window, 'same')
      GD = GD - GD[i_center]
      GDD = np.gradient(GD) / np.gradient(2*np.pi * f_SPIDER_shifted)
      phase = scipy.integrate.cumtrapz(GD, f_SPIDER_shifted * 2 * np.pi, initial = 0.0)
      phase = phase - phase[i_center]
      t, I_t = calculate_pulse(lambda_fundamental = lambda_fundamental, I_fundamental = I_fundamental, f_SPIDER_shifted = f_SPIDER_shifted, phase = phase)
      I_t = I_t / max(abs(I_t))
      I_fundamental_reconstructed = abs(E_SPIDER_AC)**2
      return f_SPIDER_shifted * 1e-12, I_fundamental_reconstructed * 1e12, phase, GD * 1e15, GDD * 1e30, t * 1e15, I_t * 1e-15

###############################################################################
"""
Multilayer Optics
"""

def multilayer_transfermatrix(n_array, d_array, lambda0, theta_in):
    """
    returns r, t, theta_out (rad)
    lambda0 = vacuum wavelength
    n_array also contains the two half-space on both side of the stack. A single
    layer therefore has n_array = [1, n, 1] where n might be complex. The first 
    and last element of d_array has to be filled but is irrelevant.   
    d_array in meter
    theta_in in rad
    r = reflectivity (Reflectance = |r|^2)
    t = transmittivity (Transmittance = |t|^2)
    """
    if len(n_array) != len(d_array):
        if len(n_array) == len(d_array) + 2:
            d_array = [0] + list(d_array) + [0]
        else:
            raise ValueError('multilayer_transfermatrix: n_array and d_array need to have equal length although the first and last element of d_array are not applied.')
    M = np.array([[1, 0], [0, 1]])
    q1 =  2*np.pi / lambda0 * n_array[0] * np.cos(theta_in)
    qi = q1
    theta_i = theta_in
    for i in range(1, len(n_array)-1):
        theta_i = np.arcsin(n_array[i-1]/n_array[i] * np.sin(theta_i))
        qi = 2*np.pi / lambda0 * n_array[i] * np.cos(np.arcsin(n_array[i-1]/n_array[i] * 2*np.pi / lambda0 * np.sin(theta_i)))
        Mi = np.array([[np.cos(qi * d_array[i]), 1/qi * np.sin(qi * d_array[i])], [-qi * np.sin(qi * d_array[i]), np.cos(qi * d_array[i])]])
        M = np.dot(Mi, M)
    theta_out = np.arcsin(n_array[-2]/n_array[-1] * np.sin(theta_i))
    q2 = 2*np.pi / lambda0 * n_array[-1] * np.cos(theta_out)
    r = (q1 * q2 * M[0, 1] + M[1, 0] + 1j * (q1 * M[1, 1] - q2 * M[0, 0])) / (q1 * q2 * M[0, 1] - M[1, 0] + 1j * (q1 * M[1,1] + q2 * M[0,0]))
    t = np.exp(-1j * q2 * np.sum(d_array[1:-1])) * 2j * q1 / (q1 * q2 * M[0,1] - M[1,0] + 1j * (q1 * M[1,1] + q2 * M[0,0]))
    return r, t, theta_out


    